# Generated by Django 3.2.20 on 2023-09-14 09:50

from django.db import migrations


def set_journal_entry_period(apps, schema_editor):
    JournalEntry = apps.get_model("leprikon", "JournalEntry")
    SchoolYearPeriod = apps.get_model("leprikon", "SchoolYearPeriod")
    for entry in JournalEntry.objects.filter(
        journal__school_year_division__isnull=False, period__isnull=True
    ).select_related("journal"):
        entry.period = (
            SchoolYearPeriod.objects.filter(
                school_year_division_id=entry.journal.school_year_division_id,
            )
            .exclude(
                start__gt=entry.date,
                end__lt=entry.date,
            )
            .first()
            or SchoolYearPeriod.objects.filter(
                school_year_division_id=entry.journal.school_year_division_id,
            ).first()
        )
        entry.save()


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0081_min_due_date_days"),
    ]

    operations = [
        migrations.RunPython(set_journal_entry_period),
    ]



import re
from typing import Any, Optional

from cms.forms.fields import PageSelectFormField
from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.db import models
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _
from localflavor.cz.forms import CZPostalCodeField

from ..conf import settings
from ..utils import get_birth_date
from ..utils.calendar import DayOfWeek, DaysOfWeek
from .utils import BankAccount, parse_bank_account


class ColorInput(forms.TextInput):
    input_type = "color"


class ColorField(models.CharField):
    default_validators = [
        RegexValidator(
            re.compile("^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$"),
            _("Enter a valid hex color."),
            "invalid",
        )
    ]

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 10
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        kwargs["widget"] = ColorInput
        return super().formfield(**kwargs)


class EmailField(models.EmailField):
    def to_python(self, value):
        v = super().to_python(value)
        return v and v.lower()


class PriceField(models.DecimalField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("decimal_places", settings.PRICE_DECIMAL_PLACES)
        kwargs.setdefault("max_digits", settings.PRICE_MAX_DIGITS)
        super().__init__(*args, **kwargs)


class DaysOfWeekField(models.SmallIntegerField):
    """
    A field for storing a list of days of the week as a bitmask.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validators = ()

    def from_db_value(self, value: Optional[int], expression, connection) -> Optional[DaysOfWeek]:
        return None if value is None else DaysOfWeek(value)

    def to_python(self, value: Any) -> Optional[DaysOfWeek]:
        return None if value is None else DaysOfWeek(value)

    def get_prep_value(self, value: Any) -> Optional[int]:
        return None if value is None else sum(DaysOfWeek(value))

    def formfield(self, widget=None, form_class=None, choices_form_class=None, **kwargs):
        """Return a django.forms.Field instance for this field."""

        defaults = {
            "required": not self.blank,
            "label": capfirst(self.verbose_name),
            "choices": DayOfWeek.choices,
            "coerce": int,
            "widget": forms.SelectMultiple,
            "help_text": self.help_text,
        }
        if self.has_default():
            if callable(self.default):
                defaults["initial"] = self.default
                defaults["show_hidden_initial"] = True
            else:
                defaults["initial"] = self.get_default()
        if self.null:
            defaults["empty_value"] = None
        if choices_form_class is not None:
            form_class = choices_form_class
        else:
            form_class = forms.TypedMultipleChoiceField
        # Many of the subclass-specific formfield arguments (min_value,
        # max_value) don't apply for choice fields, so be sure to only pass
        # the values that TypedChoiceField will understand.
        for k in list(kwargs):
            if k not in (
                "coerce",
                "empty_value",
                "choices",
                "required",
                "widget",
                "label",
                "initial",
                "help_text",
                "error_messages",
                "show_hidden_initial",
                "disabled",
            ):
                del kwargs[k]
        defaults.update(kwargs)
        return form_class(**defaults)


@DaysOfWeekField.register_lookup
class BitwiseMatchAny(models.Lookup):
    lookup_name = "match"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        return f"({lhs} & {rhs}) <> 0", lhs_params + rhs_params


birth_num_regex = re.compile("^[0-9]{2}([0257][1-9]|[1368][0-2])[0-3][0-9]/?[0-9]{3,4}$")


def validate_birth_num(value):
    try:
        assert birth_num_regex.match(value) is not None
        get_birth_date(value)
        value = value.replace("/", "")
        if len(value) > 9:
            assert int(value) % 11 == 0
    except Exception:
        raise ValidationError(
            message=_("Enter a valid birth number."),
            code="invalid",
        )


class BirthNumberFormField(forms.CharField):
    default_validators = [validate_birth_num]


class BirthNumberField(models.CharField):
    default_validators = [validate_birth_num]

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 11
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if kwargs["max_length"] == 11:
            del kwargs["max_length"]
        return name, path, args, kwargs

    def clean(self, value, model_instance):
        value = super().clean(value, model_instance)
        return value if value[6] == "/" else "{}/{}".format(value[:6], value[6:])

    def formfield(self, **kwargs):
        kwargs.setdefault("form_class", BirthNumberFormField)
        return super().formfield(**kwargs)


class BankAccountFormField(forms.CharField):
    validators = [parse_bank_account]


class BankAccountField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("max_length", 34)
        super().__init__(*args, **kwargs)
        self.validators = [parse_bank_account]

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if kwargs["max_length"] == 34:
            del kwargs["max_length"]
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        kwargs.setdefault("form_class", BankAccountFormField)
        return super().formfield(**kwargs)

    def from_db_value(self, value: Optional[str], expression, connection):
        if value is None:
            return value
        return BankAccount(value)

    def get_prep_value(self, value: Optional[BankAccount]) -> Optional[str]:
        return value and value.iban.compact

    def to_python(self, value: Any) -> Optional[BankAccount]:
        if value is None:
            return value
        return parse_bank_account(value)


class _PostalCodeField(CZPostalCodeField, forms.CharField):
    """
    CZPostalCodeField derived from CharField instead of just Field
    to support max_length
    """


class PostalCodeField(models.CharField):
    def __init__(self, *args, **kwargs):
        defaults = {"max_length": 6}
        defaults.update(kwargs)
        super().__init__(*args, **defaults)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_length"]
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        defaults = {"form_class": _PostalCodeField}
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def clean(self, value, model_instance):
        return super().clean(
            value[3] == " " and value or "{} {}".format(value[:3], value[3:]),
            model_instance,
        )


class UniquePageField(models.OneToOneField):
    """
    This is a copy of PageField based on OneToOneField instead of ForeignKey.
    """

    default_form_class = PageSelectFormField

    def __init__(self, **kwargs):
        # We hard-code the `to` argument for ForeignKey.__init__
        # since a PageField can only be a ForeignKey to a Page
        kwargs["to"] = "cms.Page"
        kwargs["on_delete"] = kwargs.get("on_delete", models.CASCADE)
        super().__init__(**kwargs)

    def formfield(self, **kwargs):
        defaults = {
            "form_class": self.default_form_class,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)



from adminsortable2.admin import SortableAdminMixin
from django.contrib import admin

from ..models.statgroup import StatGroup


@admin.register(StatGroup)
class StatGroupAdmin(SortableAdminMixin, admin.ModelAdmin):
    list_display = ("name",)



# Generated by Django 3.2.25 on 2024-05-09 08:48

from django.db import migrations
import leprikon.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0083_statgroup"),
    ]

    operations = [
        migrations.AddField(
            model_name="subjectregistration",
            name="cached_balance",
            field=leprikon.models.fields.PriceField(
                decimal_places=0, default=0, editable=False, max_digits=10, verbose_name="payments balance"
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 1.10.7 on 2017-09-04 19:32

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("leprikon", "0002_initial_data"),
    ]

    operations = [
        migrations.CreateModel(
            name="Manager",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "school_years",
                    models.ManyToManyField(
                        related_name="managers", to="leprikon.SchoolYear", verbose_name="school years"
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="leprikon_manager",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "ordering": ("user__first_name", "user__last_name"),
                "verbose_name": "manager",
                "verbose_name_plural": "managers",
            },
        ),
        migrations.AddField(
            model_name="courselistplugin",
            name="managers",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by managers.",
                related_name="+",
                to="leprikon.Manager",
                verbose_name="managers",
            ),
        ),
        migrations.AddField(
            model_name="eventlistplugin",
            name="managers",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by managers.",
                related_name="+",
                to="leprikon.Manager",
                verbose_name="managers",
            ),
        ),
        migrations.AddField(
            model_name="subject",
            name="managers",
            field=models.ManyToManyField(
                blank=True, related_name="subjects", to="leprikon.Manager", verbose_name="managers"
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2022-02-20 19:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0067_remove_rocketchat"),
    ]

    operations = [
        migrations.AddField(
            model_name="participant",
            name="answers",
            field=models.TextField(blank=True, default="{}", editable=False, verbose_name="additional answers"),
        ),
    ]



from adminsortable2.admin import SortableAdminMixin
from django.contrib import admin

from ..models.targetgroup import TargetGroup


@admin.register(TargetGroup)
class TargetGroupAdmin(SortableAdminMixin, admin.ModelAdmin):
    list_display = ("name", "stat_group", "require_school")
    list_editable = ("require_school", "stat_group")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-09-20 21:11
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import leprikon.models.fields
import leprikon.models.activities


def set_subject_registration_type(apps, schema_editor):
    Subject = apps.get_model("leprikon", "Subject")
    Subject.objects.filter(max_group_members_count__gt=0).update(
        registration_type="G",
        min_participants_count=0,
        max_participants_count=0,
    )


def create_subject_registration_groups(apps, schema_editor):
    SubjectRegistration = apps.get_model("leprikon", "SubjectRegistration")
    SubjectRegistrationGroup = apps.get_model("leprikon", "SubjectRegistrationGroup")
    SubjectRegistrationGroup.objects.bulk_create(
        SubjectRegistrationGroup(
            registration=registration,
            target_group_id=registration.target_group_id,
            name=registration.group_name,
            first_name=registration.group_leader_first_name,
            last_name=registration.group_leader_last_name,
            street=registration.group_leader_street,
            city=registration.group_leader_city,
            postal_code=registration.group_leader_postal_code,
            phone=registration.group_leader_phone,
            email=registration.group_leader_email,
            school_id=registration.group_school_id,
            school_other=registration.group_school_other,
            school_class=registration.group_school_class,
        )
        for registration in SubjectRegistration.objects.filter(subject__registration_type="G")
    )


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0033_target_groups"),
    ]

    operations = [
        migrations.AlterField(
            model_name="subject",
            name="age_groups",
            field=models.ManyToManyField(related_name="subjects", to="leprikon.AgeGroup", verbose_name="age groups"),
        ),
        migrations.AlterField(
            model_name="subject",
            name="target_groups",
            field=models.ManyToManyField(
                related_name="subjects", to="leprikon.TargetGroup", verbose_name="target groups"
            ),
        ),
        migrations.AlterField(
            model_name="subject",
            name="max_group_members_count",
            field=models.PositiveIntegerField(
                help_text="Group member details only include name and note.",
                verbose_name="maximal group members count per registration",
            ),
        ),
        migrations.AlterField(
            model_name="subject",
            name="min_group_members_count",
            field=models.PositiveIntegerField(
                help_text="Group member details only include name and note.",
                verbose_name="minimal group members count per registration",
            ),
        ),
        migrations.AddField(
            model_name="subject",
            name="registration_type",
            field=models.CharField(
                choices=[("P", "participants"), ("G", "groups")],
                default="P",
                max_length=1,
                help_text="Participant details include birth number (birth day), age group, contacts, parent, etc. Group member details only include name and note.",
                verbose_name="registration type",
            ),
            preserve_default=False,
        ),
        migrations.RunPython(set_subject_registration_type),
        migrations.CreateModel(
            name="SubjectRegistrationGroup",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "registration",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="group",
                        to="leprikon.SubjectRegistration",
                        verbose_name="registration",
                    ),
                ),
                (
                    "target_group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="leprikon.TargetGroup",
                        verbose_name="target group",
                    ),
                ),
                ("name", models.CharField(blank=True, null=True, max_length=150, verbose_name="group name")),
                ("first_name", models.CharField(max_length=30, verbose_name="first name")),
                ("last_name", models.CharField(max_length=30, verbose_name="last name")),
                ("street", models.CharField(max_length=150, verbose_name="street")),
                ("city", models.CharField(max_length=150, verbose_name="city")),
                ("postal_code", leprikon.models.fields.PostalCodeField(verbose_name="postal code")),
                ("phone", models.CharField(max_length=30, verbose_name="phone")),
                ("email", leprikon.models.fields.EmailField(max_length=254, verbose_name="email address")),
                (
                    "school",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="leprikon.School",
                        verbose_name="school",
                    ),
                ),
                ("school_other", models.CharField(blank=True, default="", max_length=150, verbose_name="other school")),
                ("school_class", models.CharField(blank=True, default="", max_length=30, verbose_name="class")),
                (
                    "answers",
                    models.TextField(blank=True, default="{}", editable=False, verbose_name="additional answers"),
                ),
            ],
            options={
                "verbose_name": "registered group",
                "verbose_name_plural": "registered groups",
            },
            bases=(leprikon.models.activities.SchoolMixin, leprikon.models.activities.PersonMixin, models.Model),
        ),
        migrations.RunPython(create_subject_registration_groups),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_age_group",
            new_name="age_group",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_birth_num",
            new_name="birth_num",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_citizenship",
            new_name="citizenship",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_city",
            new_name="city",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_email",
            new_name="email",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_first_name",
            new_name="first_name",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_gender",
            new_name="gender",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_health",
            new_name="health",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_last_name",
            new_name="last_name",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_phone",
            new_name="phone",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_postal_code",
            new_name="postal_code",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_school",
            new_name="school",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_school_class",
            new_name="school_class",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_school_other",
            new_name="school_other",
        ),
        migrations.RenameField(
            model_name="subjectregistrationparticipant",
            old_name="participant_street",
            new_name="street",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_leader_city",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_leader_email",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_leader_first_name",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_leader_last_name",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_leader_phone",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_leader_postal_code",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_leader_street",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_name",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_school",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_school_class",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="group_school_other",
        ),
        migrations.RemoveField(
            model_name="subjectregistration",
            name="target_group",
        ),
    ]



# Generated by Django 1.11.12 on 2018-08-29 13:30

from django.db import migrations, models
from django.utils import translation
from django.utils.translation import gettext as _
from django_countries.data import COUNTRIES
import django.db.models.deletion
from leprikon.conf import settings


_("Czechia")  # we want Czechia to be translated
all_EU_countries = [
    "AT",
    "BE",
    "BG",
    "CY",
    "CZ",
    "DK",
    "EE",
    "FI",
    "FR",
    "DE",
    "GR",
    "HU",
    "IE",
    "IT",
    "LV",
    "LT",
    "LU",
    "MT",
    "NL",
    "PL",
    "PT",
    "RO",
    "SK",
    "SI",
    "ES",
    "SE",
    "GB",
]
other_EU_countries = [country for country in all_EU_countries if country != settings.LEPRIKON_COUNTRY]


def create_default_citizenships(apps, schema_editor):
    Citizenship = apps.get_model("leprikon", "Citizenship")
    translation.activate(settings.LEPRIKON_COUNTRY)
    Citizenship.objects.bulk_create(
        (
            Citizenship(order=0, name=COUNTRIES[settings.LEPRIKON_COUNTRY], require_birth_num=True),
            Citizenship(order=1, name=_("other country in EU"), require_birth_num=False),
            Citizenship(order=2, name=_("country outside EU"), require_birth_num=False),
        )
    )


def update_registration_citizenships(apps, schema_editor):
    Citizenship = apps.get_model("leprikon", "Citizenship")
    SubjectRegistration = apps.get_model("leprikon", "SubjectRegistration")
    SubjectRegistration.objects.filter(
        participant_citizenship_old=settings.LEPRIKON_COUNTRY,
    ).update(participant_citizenship_id=1)
    SubjectRegistration.objects.filter(
        participant_citizenship_old__in=other_EU_countries,
    ).update(participant_citizenship_id=2)


def update_participant_citizenships(apps, schema_editor):
    Citizenship = apps.get_model("leprikon", "Citizenship")
    Participant = apps.get_model("leprikon", "Participant")
    Participant.objects.filter(
        citizenship_old=settings.LEPRIKON_COUNTRY,
    ).update(citizenship_id=1)
    Participant.objects.filter(
        citizenship_old__in=other_EU_countries,
    ).update(citizenship_id=2)


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0016_school_year_divisions_cleanup"),
    ]

    operations = [
        migrations.CreateModel(
            name="Citizenship",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=50, verbose_name="name")),
                ("require_birth_num", models.BooleanField(default=True, verbose_name="require birth number")),
                ("order", models.IntegerField(blank=True, default=0, verbose_name="order")),
            ],
            options={
                "ordering": ("order",),
                "verbose_name": "citizenship",
                "verbose_name_plural": "citizenships",
            },
        ),
        migrations.RunPython(create_default_citizenships),
        migrations.RenameField(
            model_name="subjectregistration",
            old_name="participant_citizenship",
            new_name="participant_citizenship_old",
        ),
        migrations.AddField(
            model_name="subjectregistration",
            name="participant_citizenship",
            field=models.ForeignKey(
                default=3,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="leprikon.Citizenship",
                verbose_name="citizenship",
            ),
            preserve_default=False,
        ),
        migrations.RunPython(update_registration_citizenships),
        migrations.RenameField(
            model_name="participant",
            old_name="citizenship",
            new_name="citizenship_old",
        ),
        migrations.AddField(
            model_name="participant",
            name="citizenship",
            field=models.ForeignKey(
                default=3,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="leprikon.Citizenship",
                verbose_name="citizenship",
            ),
            preserve_default=False,
        ),
        migrations.RunPython(update_participant_citizenships),
    ]



# Generated by Django 1.9.13 on 2018-04-03 20:12

from django.db import migrations, models
import django.db.models.deletion
import djangocms_text_ckeditor.fields
import leprikon.models.fields


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0011_subjectregistration_payment_requested"),
    ]

    operations = [
        migrations.CreateModel(
            name="SubjectVariant",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=150, verbose_name="variant name")),
                (
                    "description",
                    djangocms_text_ckeditor.fields.HTMLField(
                        blank=True, default="", verbose_name="variant description"
                    ),
                ),
                (
                    "price",
                    leprikon.models.fields.PriceField(
                        blank=True, decimal_places=0, max_digits=10, null=True, verbose_name="price"
                    ),
                ),
                ("order", models.IntegerField(blank=True, default=0, verbose_name="order")),
                (
                    "subject",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="variants",
                        to="leprikon.Subject",
                        verbose_name="subject",
                    ),
                ),
            ],
            options={
                "ordering": ("subject", "order"),
                "verbose_name": "price and registering variant",
                "verbose_name_plural": "price and registering variants",
            },
        ),
        migrations.AddField(
            model_name="subjectregistration",
            name="subject_variant",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="registrations",
                to="leprikon.SubjectVariant",
                verbose_name="variant",
            ),
        ),
    ]



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext as _

from ..models.roles import FilteredLeaderListPlugin, LeaderListPlugin, LeaderPlugin
from .base import LeprikonPluginBase


@plugin_pool.register_plugin
class LeprikonLeaderPlugin(LeprikonPluginBase):
    name = _("Leader")
    model = LeaderPlugin
    raw_id_fields = ("leader",)

    def get_render_template(self, context, instance, placeholder):
        return "leprikon/cms/leader/%s.html" % instance.template


@plugin_pool.register_plugin
class LeprikonLeaderListPlugin(LeprikonPluginBase):
    name = _("Leader list")
    model = LeaderListPlugin

    def get_render_template(self, context, instance, placeholder):
        return "leprikon/cms/leader_list/%s.html" % instance.template


@plugin_pool.register_plugin
class LeprikonFilteredLeaderListPlugin(LeprikonPluginBase):
    name = _("Leader list with search form")
    model = FilteredLeaderListPlugin
    render_template = "leprikon/cms/leader_list_filtered.html"



from django.shortcuts import get_object_or_404, redirect

from ..models.activities import Activity
from ..models.registrationlink import RegistrationLink
from .activities import ActivityListBaseView, ActivityTypeMixin, RegistrationFormBaseView


class RegistrationLinkMixin(ActivityTypeMixin):
    def dispatch(self, request, registration_link, *args, **kwargs):
        self.registration_link = get_object_or_404(RegistrationLink, slug=registration_link)
        self.activity_type = self.registration_link.activity_type
        self.model = self._models[self.activity_type.model]
        return super(ActivityTypeMixin, self).dispatch(request, *args, **kwargs)

    def get_placeholder(self):
        return super().get_placeholder() + ":" + self.activity_type.slug

    def get_queryset(self):
        return super().get_queryset().filter(id__in=self.registration_link.activity_variants.values("activity_id"))

    def get_template_names(self):
        return [
            "leprikon/{}{}.html".format(self.activity_type.slug, self.template_name_suffix),
            "leprikon/{}{}.html".format(self.activity_type.model, self.template_name_suffix),
            "leprikon/activity{}.html".format(self.template_name_suffix),
        ]


class RegistrationLinkView(RegistrationLinkMixin, ActivityListBaseView):
    pass


class ActivityMixin:
    registration_link: RegistrationLink

    def dispatch(self, request, pk, **kwargs):
        lookup_kwargs = {"activity_type_id": self.registration_link.activity_type_id, "id": int(pk)}
        if not self.request.user.is_staff:
            lookup_kwargs["public"] = True
        self.activity = get_object_or_404(Activity, **lookup_kwargs)
        if not self.registration_link.registration_allowed:
            return redirect(self.activity)
        self.request.school_year = self.activity.school_year
        return super().dispatch(request, **kwargs)


class RegistrationLinkFormView(RegistrationLinkMixin, ActivityMixin, RegistrationFormBaseView):
    def form_valid(self, form):
        response = super().form_valid(form)
        self.object.registration_link = self.registration_link
        self.object.save()
        return response



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext as _

from ..models.schoolyear import SchoolYearBlockPlugin
from .base import LeprikonPluginBase


@plugin_pool.register_plugin
class LeprikonSchoolYearBlockPlugin(LeprikonPluginBase):
    name = _("School year block")
    model = SchoolYearBlockPlugin
    allow_children = True
    filter_horizontal = ("school_years",)
    render_template = "leprikon/cms/school_year_block.html"



# Generated by Django 3.2.20 on 2023-08-16 20:04

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0079_mandatory_subject_variants"),
    ]

    operations = [
        migrations.DeleteModel(
            name="CourseRegistrationHistory",
        ),
    ]



from django import forms
from django.contrib.auth import get_user_model, password_validation
from django.contrib.auth.forms import (
    AuthenticationForm as _AuthenticationForm,
    PasswordChangeForm as _PasswordChangeForm,
    PasswordResetForm as _PasswordResetForm,
    SetPasswordForm as _SetPasswordForm,
    UserChangeForm as _UserChangeForm,
    UserCreationForm as _UserCreationForm,
)
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from verified_email_field.auth import EmailAuthenticationForm as _EmailAuthenticationForm
from verified_email_field.forms import VerifiedEmailField

from .form import FormMixin

User = get_user_model()


class UserFormMixin(FormMixin):
    def clean_email(self):
        email = self.cleaned_data["email"].lower()
        user_qs = User.objects.filter(email=email)
        if self.instance.pk:
            user_qs = user_qs.exclude(pk=self.instance.pk)
        if email and user_qs.exists():
            raise ValidationError(
                _("User with this email already exists."),
                code="exists",
                params={"email": email},
            )
        else:
            return email


class UserAdminCreateForm(UserFormMixin, forms.ModelForm):
    # create user with random password
    # users without password may not reset the password themselves
    def save(self, commit=True):
        user = super().save(commit)
        user.set_password(User.objects.make_random_password())
        user.save()
        return user

    class Meta:
        model = User
        fields = ["username", "first_name", "last_name", "email"]


class UserAdminChangeForm(UserFormMixin, _UserChangeForm):
    pass


class UserCreateForm(UserFormMixin, _UserCreationForm):
    email = VerifiedEmailField(label=_("E-mail"), fieldsetup_id="UserCreateForm", required=True)
    agreement = forms.BooleanField(label=_("I agree with the terms."), required=True)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.fields["username"].help_text = None
        self.fields["password1"].help_text = password_validation.password_validators_help_text_html()

    class Meta:
        model = User
        fields = ["username", "first_name", "last_name", "email"]


class UserAgreementForm(FormMixin, forms.Form):
    agreement = forms.BooleanField(label=_("I agree with the terms."), required=True)


class UserEmailForm(UserFormMixin, forms.ModelForm):
    email = VerifiedEmailField(label=_("E-mail"), fieldsetup_id="UserEmailForm", required=True)

    class Meta:
        model = User
        fields = ["email"]


class UserUpdateForm(UserFormMixin, forms.ModelForm):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.fields["username"].help_text = None

    class Meta:
        model = User
        fields = ["username", "first_name", "last_name"]


class UserLoginForm:
    class PasswordForm(FormMixin, _AuthenticationForm):
        pass

    class EmailForm(FormMixin, _EmailAuthenticationForm):
        pass

    def __init__(self, request, data=None, **kwargs):
        if data:
            if "email_0" in data:
                self.used_form = self.email_form = self.EmailForm(data=data)
                self.password_form = self.PasswordForm(request=request)
            else:
                self.used_form = self.password_form = self.PasswordForm(data=data, request=request)
                self.email_form = self.EmailForm()
        else:
            self.password_form = self.PasswordForm(request=request)
            self.email_form = self.EmailForm()
            self.used_form = None
        self.media = self.password_form.media + self.email_form.media

    def __getattr__(self, attr):
        # implement an interface to the currently used form
        return getattr(self.used_form, attr)


class UserPasswordForm(FormMixin, _PasswordChangeForm):
    pass


class PasswordResetForm(FormMixin, _PasswordResetForm):
    def get_users(self, email):
        return User.objects.filter(
            **{
                "email__iexact": email,
                "is_active": True,
            }
        )


class SetPasswordForm(FormMixin, _SetPasswordForm):
    pass



from django.db import models
from django.utils.translation import gettext_lazy as _


class AccountClosureManager(models.Manager):
    def max_closure_date(self):
        return self.aggregate(closure_date=models.Max("closure_date"))["closure_date"]


class AccountClosure(models.Model):
    closure_date = models.DateField(_("closure date"), db_index=True)

    objects = AccountClosureManager()

    class Meta:
        app_label = "leprikon"
        ordering = ("-closure_date",)
        verbose_name = _("account closure")
        verbose_name_plural = _("account closures")

    def __str__(self):
        return str(self.closure_date)



# Generated by Django 3.2.22 on 2023-11-08 09:26

from django.db import migrations, models


default_stat_group = None


def create_default_stat_group(apps, schema_editor):
    global default_stat_group
    StatGroup = apps.get_model("leprikon", "StatGroup")
    default_stat_group = StatGroup.objects.create(name="výchozí")


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0082_journal_entry_period"),
    ]

    operations = [
        migrations.CreateModel(
            name="StatGroup",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=50, verbose_name="name")),
                ("order", models.IntegerField(blank=True, default=0, verbose_name="order")),
            ],
            options={
                "verbose_name": "group for statistics",
                "verbose_name_plural": "groups for statistics",
                "ordering": ("order",),
            },
        ),
        migrations.RunPython(create_default_stat_group),
        migrations.AddField(
            model_name="agegroup",
            name="stat_group",
            field=models.ForeignKey(
                default=lambda: default_stat_group.id,
                on_delete=models.deletion.PROTECT,
                related_name="+",
                to="leprikon.statgroup",
                verbose_name="group for statistics",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="targetgroup",
            name="stat_group",
            field=models.ForeignKey(
                default=lambda: default_stat_group.id,
                on_delete=models.deletion.PROTECT,
                related_name="+",
                to="leprikon.statgroup",
                verbose_name="group for statistics",
            ),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-02-11 04:57
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import djangocms_text_ckeditor.fields


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0024_subjectregistration_note"),
    ]

    operations = [
        migrations.CreateModel(
            name="Agreement",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=50, unique=True, verbose_name="name")),
                ("order", models.IntegerField(blank=True, default=0, verbose_name="order")),
                ("heading", models.CharField(blank=True, default="", max_length=50, verbose_name="heading")),
                (
                    "agreement",
                    djangocms_text_ckeditor.fields.HTMLField(blank=True, default="", verbose_name="agreement"),
                ),
            ],
            options={
                "verbose_name": "legal agreement",
                "verbose_name_plural": "legal agreements",
                "ordering": ("order",),
            },
        ),
        migrations.CreateModel(
            name="AgreementOption",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=50, verbose_name="name")),
                ("order", models.IntegerField(blank=True, default=0, verbose_name="order")),
                ("required", models.BooleanField(default=False, verbose_name="required")),
                ("option", djangocms_text_ckeditor.fields.HTMLField(verbose_name="option")),
                (
                    "agreement",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="options",
                        to="leprikon.Agreement",
                        verbose_name="agreement",
                    ),
                ),
            ],
            options={
                "verbose_name": "agreement option",
                "verbose_name_plural": "agreement options",
                "ordering": ("order",),
            },
        ),
        migrations.RenameField(
            model_name="leprikonsite",
            old_name="registration_agreement",
            new_name="old_registration_agreement",
        ),
        migrations.RenameField(
            model_name="subject",
            old_name="agreement",
            new_name="old_registration_agreement",
        ),
        migrations.RenameField(
            model_name="subjecttype",
            old_name="agreement",
            new_name="old_registration_agreement",
        ),
        migrations.AlterField(
            model_name="leprikonsite",
            name="old_registration_agreement",
            field=djangocms_text_ckeditor.fields.HTMLField(
                blank=True,
                default="",
                help_text="This agreement will be removed in future version. Please, use registration agreements below.",
                verbose_name="old registration agreement",
            ),
        ),
        migrations.AlterField(
            model_name="subject",
            name="old_registration_agreement",
            field=djangocms_text_ckeditor.fields.HTMLField(
                blank=True,
                default="",
                help_text="This agreement will be removed in future version. Please, use registration agreements below.",
                verbose_name="old registration agreement",
            ),
        ),
        migrations.AlterField(
            model_name="subjecttype",
            name="old_registration_agreement",
            field=djangocms_text_ckeditor.fields.HTMLField(
                blank=True,
                default="",
                help_text="This agreement will be removed in future version. Please, use registration agreements below.",
                verbose_name="old registration agreement",
            ),
        ),
        migrations.AddField(
            model_name="leprikonsite",
            name="registration_agreements",
            field=models.ManyToManyField(
                blank=True,
                help_text="Add legal agreements for the registration form.",
                related_name="+",
                to="leprikon.Agreement",
                verbose_name="registration agreements",
            ),
        ),
        migrations.AddField(
            model_name="subject",
            name="registration_agreements",
            field=models.ManyToManyField(
                blank=True,
                help_text="Add additional legal agreements specific for this subject.",
                related_name="+",
                to="leprikon.Agreement",
                verbose_name="additional legal agreements",
            ),
        ),
        migrations.AddField(
            model_name="subjectregistration",
            name="agreement_options",
            field=models.ManyToManyField(blank=True, to="leprikon.AgreementOption", verbose_name="legal agreements"),
        ),
        migrations.AddField(
            model_name="subjecttype",
            name="registration_agreements",
            field=models.ManyToManyField(
                blank=True,
                help_text="Add additional legal agreements specific to this activity type.",
                related_name="+",
                to="leprikon.Agreement",
                verbose_name="additional legal agreements",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="agreementoption",
            unique_together=set([("agreement", "name")]),
        ),
    ]



# Generated by Django 3.2.20 on 2023-07-27 11:28

from django.db import migrations, models
import django.db.models.deletion
import leprikon.models.fields


def update_current_subject_variants(apps, schema_editor):
    Subject = apps.get_model("leprikon", "Subject")
    for subject in Subject.objects.all():
        subject.variants.filter(available=True).update(reg_from=subject.reg_from, reg_to=subject.reg_to)
        for attr in [
            "min_participants_count",
            "max_participants_count",
            "registration_price",
            "participant_price",
        ]:
            subject.variants.filter(**{attr: None}).update(**{attr: getattr(subject, attr)})


def create_default_subject_variants(apps, schema_editor):
    Subject = apps.get_model("leprikon", "Subject")
    SubjectVariant = apps.get_model("leprikon", "SubjectVariant")
    SubjectRegistration = apps.get_model("leprikon", "SubjectRegistration")

    subject_variants = {
        subject_variant.subject_id: subject_variant
        for subject_variant in SubjectVariant.objects.bulk_create(
            SubjectVariant(
                subject=subject,
                name="",
                description="",
                reg_from=subject.reg_from,
                reg_to=subject.reg_to,
                min_participants_count=subject.min_participants_count,
                max_participants_count=subject.max_participants_count,
                registration_price=subject.registration_price,
                participant_price=subject.participant_price,
                school_year_division_id=subject.school_year_division_id,
                allow_period_selection=subject.allow_period_selection or False,
            )
            for subject in Subject.objects.annotate(
                num_variants=models.Count("variants"),
                school_year_division_id=models.F("course__school_year_division_id"),
                allow_period_selection=models.F("course__allow_period_selection"),
            )
            if subject.num_variants == 0
        )
    }

    for subject_variant in subject_variants.values():
        subject_variant.age_groups.set(subject_variant.subject.age_groups.all())
        subject_variant.target_groups.set(subject_variant.subject.target_groups.all())

        SubjectRegistration.objects.filter(subject=subject_variant.subject, subject_variant=None).update(
            subject_variant=subject_variant
        )

    for reg in (
        SubjectRegistration.objects.filter(subject_variant=None)
        .annotate(
            school_year_division_id=models.F("subject__course__school_year_division_id"),
            allow_period_selection=models.F("subject__course__allow_period_selection"),
        )
        .select_related("subject")
    ):
        if reg.subject_id not in subject_variants:
            subject_variants[reg.subject_id] = SubjectVariant.objects.create(
                subject=reg.subject,
                name="",
                description="",
                reg_from=reg.subject.reg_from,
                reg_to=reg.subject.reg_to,
                min_participants_count=reg.subject.min_participants_count,
                max_participants_count=reg.subject.max_participants_count,
                registration_price=reg.subject.registration_price,
                participant_price=reg.subject.participant_price,
                school_year_division_id=reg.school_year_division_id,
                allow_period_selection=reg.allow_period_selection,
            )
            subject_variants[reg.subject_id].age_groups.set(reg.subject.age_groups.all())
            subject_variants[reg.subject_id].target_groups.set(reg.subject.target_groups.all())
        reg.subject_variant = subject_variants[reg.subject_id]
        reg.save()


def update_registration_links(apps, schema_editor):
    RegistrationLink = apps.get_model("leprikon", "RegistrationLink")
    SubjectVariant = apps.get_model("leprikon", "SubjectVariant")

    for registration_link in RegistrationLink.objects.all():
        registration_link.subject_variants.set(
            SubjectVariant.objects.filter(subject__in=registration_link.subjects.all())
        )


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0078_delete_coursevariant"),
    ]

    operations = [
        # migrate SubjectVariant
        migrations.AddField(
            model_name="subjectvariant",
            name="reg_from",
            field=models.DateTimeField(blank=True, null=True, verbose_name="registration active from"),
        ),
        migrations.AddField(
            model_name="subjectvariant",
            name="reg_to",
            field=models.DateTimeField(blank=True, null=True, verbose_name="registration active to"),
        ),
        migrations.RunPython(update_current_subject_variants),
        migrations.RunPython(create_default_subject_variants),
        migrations.AlterField(
            model_name="subjectregistration",
            name="subject_variant",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="registrations",
                to="leprikon.subjectvariant",
                verbose_name="variant",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvariant",
            name="max_participants_count",
            field=models.PositiveIntegerField(
                default=1, verbose_name="maximal participants / group members count per registration"
            ),
        ),
        migrations.AlterField(
            model_name="subjectvariant",
            name="min_participants_count",
            field=models.PositiveIntegerField(
                default=1, verbose_name="minimal participants / group members count per registration"
            ),
        ),
        migrations.AlterField(
            model_name="subjectvariant",
            name="name",
            field=models.CharField(blank=True, default="", max_length=150, verbose_name="variant name"),
        ),
        migrations.AlterField(
            model_name="subjectvariant",
            name="participant_price",
            field=leprikon.models.fields.PriceField(
                decimal_places=0, default=0, max_digits=10, verbose_name="price per participant"
            ),
        ),
        migrations.AlterField(
            model_name="subjectvariant",
            name="registration_price",
            field=leprikon.models.fields.PriceField(
                decimal_places=0, default=0, max_digits=10, verbose_name="price per registration"
            ),
        ),
        migrations.RemoveField(
            model_name="course",
            name="allow_period_selection",
        ),
        migrations.RemoveField(
            model_name="course",
            name="school_year_division",
        ),
        migrations.RemoveField(
            model_name="subject",
            name="max_participants_count",
        ),
        migrations.RemoveField(
            model_name="subject",
            name="min_participants_count",
        ),
        migrations.RemoveField(
            model_name="subject",
            name="participant_price",
        ),
        migrations.RemoveField(
            model_name="subject",
            name="reg_from",
        ),
        migrations.RemoveField(
            model_name="subject",
            name="reg_to",
        ),
        migrations.RemoveField(
            model_name="subject",
            name="registration_price",
        ),
        migrations.RemoveField(
            model_name="subjectvariant",
            name="available",
        ),
        migrations.RemoveField(
            model_name="courseregistration",
            name="school_year_division",
        ),
        # migrate RegistrationLink
        migrations.AddField(
            model_name="registrationlink",
            name="subject_variants",
            field=models.ManyToManyField(
                related_name="registration_links", to="leprikon.SubjectVariant", verbose_name="subjects"
            ),
        ),
        migrations.RunPython(update_registration_links),
        migrations.RemoveField(
            model_name="registrationlink",
            name="subjects",
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2022-09-04 23:34
from __future__ import unicode_literals

from decimal import Decimal
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0070_decision_print_setup"),
    ]

    operations = [
        migrations.AlterField(
            model_name="transaction",
            name="amount",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=10,
                validators=[django.core.validators.MinValueValidator(Decimal("0.01"))],
                verbose_name="amount",
            ),
        ),
    ]



import locale
import re
import unicodedata
import zlib
from datetime import date
from decimal import Decimal
from typing import Iterable, Union
from urllib.parse import parse_qs, urlencode

from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, transaction
from django.urls import reverse_lazy as reverse
from django.utils.encoding import iri_to_uri
from django.utils.functional import lazy
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ..conf import settings

MALE = "m"
FEMALE = "f"


def _get_localeconv():
    """
    This function loads localeconv during module load.
    It is necessary, because using locale.setlocale later may be dangerous
    (It is not thread-safe in most of the implementations.)
    """
    original_locale_name = locale.setlocale(locale.LC_ALL)
    locale_name = locale.locale_alias[settings.LANGUAGE_CODE].split(".")[0] + ".UTF-8"
    locale.setlocale(locale.LC_ALL, str(locale_name))
    lc = locale.localeconv()
    locale.setlocale(locale.LC_ALL, original_locale_name)
    return lc


localeconv = _get_localeconv()


def attributes(**kwargs):
    def _attributes(obj):
        for attr, value in kwargs.items():
            setattr(obj, attr, value)
        return obj

    return _attributes


def first(*values):
    for value in values:
        if value is not None:
            return value


# This function is inspired by python's standard locale.currency().
def currency(val: Union[int, float, Decimal], international=False) -> str:
    """Formats val according to the currency settings for current language."""
    digits = settings.PRICE_DECIMAL_PLACES

    # grouping
    groups = []
    s = str(abs(int(val)))
    for interval in locale._grouping_intervals(localeconv["mon_grouping"]):  # type: ignore
        if not s:
            break
        groups.append(s[-interval:])
        s = s[:-interval]
    if s:
        groups.append(s)
    groups.reverse()
    s = localeconv["mon_thousands_sep"].join(groups)

    # display fraction for non integer values
    if digits and not isinstance(val, int):
        s += localeconv["mon_decimal_point"] + "{{:.{}f}}".format(digits).format(val).split(".")[1]

    # '<' and '>' are markers if the sign must be inserted between symbol and value
    s = "<" + s + ">"

    smb = localeconv[international and "int_curr_symbol" or "currency_symbol"]
    precedes = localeconv[val < 0 and "n_cs_precedes" or "p_cs_precedes"]
    separated = localeconv[val < 0 and "n_sep_by_space" or "p_sep_by_space"]

    if precedes:
        s = smb + (separated and " " or "") + s
    else:
        s = s + (separated and " " or "") + smb

    sign_pos = localeconv[val < 0 and "n_sign_posn" or "p_sign_posn"]
    sign = localeconv[val < 0 and "negative_sign" or "positive_sign"]

    if sign_pos == 0:
        s = "(" + s + ")"
    elif sign_pos == 1:
        s = sign + s
    elif sign_pos == 2:
        s = s + sign
    elif sign_pos == 3:
        s = s.replace("<", sign)
    elif sign_pos == 4:
        s = s.replace(">", sign)
    else:
        # the default if nothing specified;
        # this should be the most fitting sign position
        s = sign + s

    return s.replace("<", "").replace(">", "").replace(" ", "\u00a0")


def amount_color(amount):
    if amount > 0:
        return settings.LEPRIKON_COLOR_POSITIVE
    elif amount < 0:
        return settings.LEPRIKON_COLOR_NEGATIVE
    else:
        return settings.LEPRIKON_COLOR_ZERO


def ascii(value):
    return unicodedata.normalize("NFKD", value).encode("ascii", errors="ignore").decode("ascii")


def comma_separated(lst: Iterable) -> str:
    lst = list(map(str, lst))
    if len(lst) > 2:
        return _(", and ").join([", ".join(lst[:-1]), lst[-1]])
    else:
        return _(", and ").join(lst)


def current_url(request):
    if request.META["QUERY_STRING"]:
        return "{}?{}".format(request.path, request.META["QUERY_STRING"])
    else:
        return request.path


def url_back(request):
    return request.POST.get(
        settings.LEPRIKON_PARAM_BACK,
        request.GET.get(
            settings.LEPRIKON_PARAM_BACK,
            reverse("leprikon:summary"),
        ),
    )


recursive_back_splitter = re.compile(f"[?&]{settings.LEPRIKON_PARAM_BACK}=")


def url_with_back(url, url_back):
    try:
        query = url_back.split("?")[1]
    except IndexError:
        pass
    else:
        try:
            # try to reuse original back url
            url_back = parse_qs(query)[settings.LEPRIKON_PARAM_BACK][0]
        except KeyError:
            pass
    # remove recursive back url
    url_back = recursive_back_splitter.split(url_back)[0]
    return "{}?{}".format(url, urlencode({settings.LEPRIKON_PARAM_BACK: iri_to_uri(url_back)}))


def reverse_with_back(request, *args, **kwargs):
    return url_with_back(reverse(*args, **kwargs), current_url(request))


def get_gender(birth_num):
    return FEMALE if birth_num[2:4] > "50" else MALE


def get_birth_date(birth_num):
    birth_num = birth_num.replace("/", "")
    y = int(birth_num[:2])
    if len(birth_num) == 9:
        # before 1954
        if y < 54:
            year = 1900 + y
        else:
            year = 1800 + y
    else:
        year = int(date.today().year / 100) * 100 + y
        if y > date.today().year % 100:
            year -= 100
    month = int(birth_num[2:4]) % 50 % 20
    day = int(birth_num[4:6])
    return date(year, month, day)


def get_age(birth_date, today=None):
    today = today or date.today()
    try:
        birth_day_this_year = date(today.year, birth_date.month, birth_date.day)
    except ValueError:
        birth_day_this_year = date(today.year, birth_date.month + 1, 1)
    if birth_day_this_year > today:
        return today.year - birth_date.year - 1
    else:
        return today.year - birth_date.year


def first_upper(s):
    return s[0].upper() + s[1:] if s else ""


def merge_objects(source, target, attributes=None, exclude=[]):
    attributes = attributes or [f.name for f in source._meta.fields if f.name not in exclude]
    for attr in attributes:
        if not getattr(target, attr):
            setattr(target, attr, getattr(source, attr))
    return target


@transaction.atomic
def merge_users(source, target):
    from .models.activities import Registration

    target = merge_objects(source, target, ("first_name", "last_name", "email"))
    target.date_joined = (
        min(source.date_joined, target.date_joined)
        if source.date_joined and target.date_joined
        else source.date_joined or target.date_joined
    )
    target.last_login = (
        max(source.last_login, target.last_login)
        if source.last_login and target.last_login
        else source.last_login or target.last_login
    )

    try:
        leader = source.leprikon_leader
        leader.user = target
        leader.save()
    except ObjectDoesNotExist:
        pass
    except IntegrityError:
        # both users are leaders
        raise

    for attr in (
        "user",
        "created_by",
        "approved_by",
        "payment_requested_by",
        "refund_offered_by",
        "cancelation_requested_by",
        "canceled_by",
    ):
        Registration.objects.filter(**{attr: source}).update(**{attr: target})

    for sp in source.leprikon_participants.all():
        tp = target.leprikon_participants.filter(birth_num=sp.birth_num).first()
        if tp:
            tp = merge_objects(sp, tp, exclude=("id", "user", "birth_num"))
            tp.save()
        else:
            sp.user = target
            sp.save()

    for sp in source.leprikon_parents.all():
        tp = target.leprikon_parents.filter(first_name=sp.first_name, last_name=sp.last_name).first()
        if tp:
            tp = merge_objects(sp, tp, exclude=("id", "user"))
            tp.save()
        else:
            sp.user = target
            sp.save()

    for sbi in source.leprikon_billing_info.all():
        tbi = target.leprikon_billing_info.filter(name=sbi.name).first()
        if tbi:
            tbi = merge_objects(sbi, tbi, exclude=("id", "user"))
            tbi.save()
        else:
            sbi.user = target
            sbi.save()

    for mr in source.leprikon_messages.all():
        if not target.leprikon_messages.filter(message=mr.message).exists():
            mr.recipient = target
            mr.save()

    try:
        # support social auth
        source.social_auth.update(user=target)
    except AttributeError:
        pass

    source.delete()


def spayd(*items):
    s = "SPD*1.0*" + "*".join(
        "%s:%s" % (k, unicodedata.normalize("NFKD", str(v).replace("*", "")).encode("ascii", "ignore").upper().decode())
        for k, v in sorted(items)
    )
    s += "*CRC32:%x" % zlib.crc32(s.encode())
    return s.upper()


def paragraph(text):
    return mark_safe(f"<p>{text.strip()}</p>".replace("\n", "<br/>\n").replace("<br/>\n<br/>\n", "</p>\n\n<p>"))


lazy_paragraph = lazy(paragraph, str)



import csv
from datetime import datetime
from functools import partial
from typing import Any, Callable, Self, Sequence

import django_excel
from django.core.exceptions import FieldDoesNotExist
from django.db.models import F, QuerySet
from django.http import HttpRequest, HttpResponse
from django.http.response import HttpResponseBase
from django.utils.translation import gettext_lazy as _

from ..utils import attributes


def get_attr_value(obj, name):
    value = getattr(obj, name)
    if callable(value):
        value = value()
    return value


def get_verbose_name(field) -> str:
    try:
        return str(field.verbose_name)
    except AttributeError:
        return str(field.related_model._meta.verbose_name)


class AdminExportMixin:
    actions: Sequence[Callable[[Self, HttpRequest, QuerySet[Any, Any]], HttpResponseBase | None] | str] | None = (
        "export_as_xlsx",
        "export_as_csv",
    )

    def get_list_export(self, request):
        try:
            return self.list_export
        except AttributeError:
            return self.list_display

    def get_export_fields(self, request):
        fields = []
        for name in self.get_list_export(request):
            try:
                names = name.split("__")
                field = self.model._meta.get_field(names[0])
                verbose_names = [get_verbose_name(field)]
                for n in names[1:]:
                    field = field.related_model._meta.get_field(n)
                    verbose_names.append(get_verbose_name(field))
                fields.append(
                    {
                        "annotate": name if len(names) > 1 else None,
                        "verbose_name": " / ".join(verbose_names),
                        "get_value": partial(lambda name, obj: get_attr_value(obj, name), name),
                    }
                )
            except (AttributeError, FieldDoesNotExist):
                if callable(name):
                    fields.append(
                        {
                            "verbose_name": getattr(name, "short_description", name.__func__.__name__),
                            "get_value": partial(lambda name, obj: name(obj), name),
                        }
                    )
                elif hasattr(self, name):
                    attr = getattr(self, name)
                    fields.append(
                        {
                            "verbose_name": getattr(attr, "short_description", name),
                            "get_value": partial(lambda attr, obj: attr(obj), attr),
                        }
                    )
                elif hasattr(self.model, name):
                    attr = getattr(self.model, name)
                    fields.append(
                        {
                            "verbose_name": getattr(attr, "short_description", name),
                            "get_value": partial(lambda name, obj: get_attr_value(obj, name), name),
                        }
                    )
                else:
                    raise Exception('Can not resolve name "{}"'.format(name))
        return fields

    def get_export_data(self, request, queryset):
        fields = self.get_export_fields(request)
        yield [str(f["verbose_name"]) for f in fields]
        annotations = {field["annotate"]: F(field["annotate"]) for field in fields if field.get("annotate")}
        for obj in queryset.annotate(**annotations):
            values = []
            for field in fields:
                value = field["get_value"](obj)
                if value is None:
                    value = ""
                if isinstance(value, datetime):
                    value = value.replace(tzinfo=None)
                else:
                    # what can not be converted to float, must be converted to string
                    try:
                        float(value)
                    except (TypeError, ValueError):
                        value = str(value)
                values.append(value)
            yield values

    @attributes(short_description=_("Export selected records as CSV"))
    def export_as_csv(self, request, queryset):
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = 'attachment; filename="{}.csv"'.format(self.model._meta.model_name)
        data = self.get_export_data(request, queryset)
        # write data to response (use all to evaluate the map generator)
        csv.writer(response).writerows(data)
        return response

    @attributes(short_description=_("Export selected records as XLSX"))
    def export_as_xlsx(self, request, queryset):
        data = self.get_export_data(request, queryset)
        response = django_excel.make_response(django_excel.pe.Sheet(data), "xlsx")
        response["Content-Disposition"] = 'attachment; filename="{}.xlsx"'.format(self.model._meta.model_name)
        return response



from django.contrib.auth import get_user_model, login as auth_login, views as auth_views
from django.urls import reverse_lazy
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from ..forms.user import (
    PasswordResetForm,
    SetPasswordForm,
    UserAgreementForm,
    UserCreateForm,
    UserEmailForm,
    UserLoginForm,
    UserPasswordForm,
    UserUpdateForm,
)
from ..models.roles import Parent
from ..models.useragreement import UserAgreement
from .generic import CreateView, FormView, UpdateView


class UserCreateView(CreateView):
    model = get_user_model()
    form_class = UserCreateForm
    title = _("Create account")
    form_item_template_name = "leprikon/user_create_form_item.html"

    def get_message(self):
        return _("User account {} has been created.").format(self.object)

    def form_valid(self, form):
        response = super().form_valid(form)

        # first user is superuser
        if self.model.objects.count() == 1:
            self.object.is_staff = True
            self.object.is_superuser = True
            self.object.save()

        # save agreement
        UserAgreement.objects.create(
            user=self.object,
            granted=now(),
        )

        # create default parent
        parent = Parent()
        parent.first_name = self.object.first_name
        parent.last_name = self.object.last_name
        parent.email = self.object.email
        parent.user = self.object
        parent.save()

        # login
        self.object.backend = "django.contrib.auth.backends.ModelBackend"
        auth_login(self.request, self.object)
        return response


class UserAgreementView(FormView):
    form_class = UserAgreementForm
    title = _("Agreement with the terms of use")
    form_item_template_name = "leprikon/user_create_form_item.html"

    def get_message(self):
        return _("The agreement has been saved.")

    def form_valid(self, form):
        try:
            self.request.user.agreement.granted = now()
            self.request.user.agreement.save()
        except UserAgreement.DoesNotExist:
            UserAgreement.objects.create(
                user=self.request.user,
                granted=now(),
            )
        return super().form_valid(form)


class UserUpdateView(UpdateView):
    model = get_user_model()
    form_class = UserUpdateForm
    title = _("Change user")

    def get_object(self):
        return self.request.user

    def get_message(self):
        return _("User account {} has been updated.").format(self.object)


class UserEmailView(UserUpdateView):
    form_class = UserEmailForm

    def get_title(self):
        if self.request.user.email:
            return _("Change e-mail address")
        else:
            return _("Set e-mail address")


class UserPasswordView(auth_views.PasswordChangeView):
    template_name = "leprikon/password.html"
    form_class = UserPasswordForm
    success_url = reverse_lazy("leprikon:summary")

    def get_context_data(self, **kwargs):
        kwargs.update(
            {
                "submit_label": _("Change password"),
                "back_label": _("Back"),
                "back_url": reverse_lazy("leprikon:summary"),
                "placeholder": "user_password",
            }
        )
        return super().get_context_data(**kwargs)


class UserLoginView(auth_views.LoginView):
    template_name = "leprikon/login.html"
    form_class = UserLoginForm
    redirect_field_name = settings.LEPRIKON_PARAM_BACK


class UserLogoutView(auth_views.LogoutView):
    next_page = settings.LOGOUT_REDIRECT_URL or "/"
    redirect_field_name = settings.LEPRIKON_PARAM_BACK


class PasswordResetView(auth_views.PasswordResetView):
    template_name = "leprikon/password_reset.html"
    form_class = PasswordResetForm
    email_template_name = "leprikon/password_reset_email.html"
    from_email = settings.SERVER_EMAIL

    def get_context_data(self, **kwargs):
        kwargs.update(
            {
                "instructions": "<p>{}<p>".format(
                    _("Enter your email address, and we'll email instructions for setting a new one.")
                ),
                "submit_label": _("Reset my password"),
                "placeholder": "password_reset",
            }
        )
        return super().get_context_data(**kwargs)

    success_url = reverse_lazy("leprikon:password_reset_done")


class PasswordResetDoneView(auth_views.PasswordResetDoneView):
    template_name = "leprikon/password_reset_done.html"


class PasswordResetConfirmView(auth_views.PasswordResetConfirmView):
    template_name = "leprikon/password_reset_confirm.html"
    form_class = SetPasswordForm

    def get_context_data(self, **kwargs):
        kwargs.update(
            {
                "instructions": "<p>{}<p>".format(
                    _("Please enter your new password twice so we can verify you typed it in correctly.")
                ),
                "submit_label": _("Set my password"),
                "placeholder": "password_set",
            }
        )
        return super().get_context_data(**kwargs)

    success_url = reverse_lazy("leprikon:password_reset_complete")


class PasswordResetCompleteView(auth_views.PasswordResetCompleteView):
    template_name = "leprikon/password_reset_complete.html"



import uuid
from datetime import date, datetime, time, timedelta
from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, Optional

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.template.loader import get_template
from django.urls import reverse
from django.utils.formats import date_format, time_format
from django.utils.translation import gettext_lazy as _
from icalendar import Calendar, Event

from leprikon.models.leprikonsite import LeprikonSite
from leprikon.utils.calendar import (
    SimpleEvent,
    TimeSlot,
    TimeSlots,
    WeeklyTimes,
    get_conflicting_timeslots,
    get_reverse_time_slots,
    get_time_slots_by_weekly_times,
)

from .fields import DaysOfWeek, DaysOfWeekField
from .roles import Leader
from .startend import StartEndMixin
from .times import WeeklyTime

if TYPE_CHECKING:
    from .activities import Activity  # Avoid circular import


class Resource(models.Model):
    name = models.CharField(_("name"), max_length=255)
    leader = models.OneToOneField(
        Leader,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="resource",
        verbose_name=_("leader"),
    )

    class Meta:
        app_label = "leprikon"
        verbose_name = _("resource")
        verbose_name_plural = _("resources")

    def __str__(self):
        return self.name

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.leader:
            self.name = str(self.leader)
        return super().save(force_insert, force_update, using, update_fields)

    @property
    def all_availabilities(self) -> list["ResourceAvailability"]:
        return list(self.availabilities.all())

    @property
    def weekly_times(self) -> WeeklyTimes:
        return (
            WeeklyTimes(availability.weekly_time for availability in self.all_availabilities)
            if self.all_availabilities
            else WeeklyTimes.unlimited()
        )


class ResourceAvailability(StartEndMixin, models.Model):
    resource = models.ForeignKey(
        Resource, on_delete=models.CASCADE, related_name="availabilities", verbose_name=_("resource")
    )
    start_date = models.DateField(_("start date"), blank=True, null=True)
    end_date = models.DateField(_("end date"), blank=True, null=True)
    days_of_week: DaysOfWeek = DaysOfWeekField(_("days of week"))
    start_time: time = models.TimeField(_("start time"), default=time(8, 0))
    end_time: time = models.TimeField(_("end time"), default=time(16, 0))

    class Meta:
        app_label = "leprikon"
        ordering = ("start_time",)
        verbose_name = _("availability")
        verbose_name_plural = _("availabilities")

    @property
    def weekly_time(self) -> WeeklyTime:
        return WeeklyTime(
            start_date=self.start_date,
            end_date=self.end_date,
            days_of_week=self.days_of_week,
            start_time=self.start_time,
            end_time=self.end_time,
        )

    def __str__(self):
        return str(self.weekly_time)

    def __and__(self, other: "ResourceAvailability") -> Optional[WeeklyTime]:
        return self.weekly_time & other.weekly_time


class ResourceGroup(models.Model):
    name = models.CharField(_("name"), max_length=255)
    resources = models.ManyToManyField(Resource, related_name="groups", verbose_name=_("resources"))

    class Meta:
        app_label = "leprikon"
        verbose_name = _("resource group")
        verbose_name_plural = _("resource groups")

    def __str__(self):
        return self.name

    @cached_property
    def all_resources(self) -> list[Resource]:
        return list(self.resources.all())


class CalendarEvent(models.Model):
    name = models.CharField(_("name"), max_length=255)
    activity: Optional["Activity"] = models.ForeignKey(
        "leprikon.Activity",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="calendar_events",
        verbose_name=_("activity"),
    )
    start_date = models.DateField(_("start date"))
    start_time = models.TimeField(_("start time"), blank=True, null=True)
    end_date = models.DateField(_("end date"))
    end_time = models.TimeField(_("end time"), blank=True, null=True)
    effective_start = models.DateTimeField(_("effective start"), editable=False)
    effective_end = models.DateTimeField(_("effective end"), editable=False)
    preparation_time = models.DurationField(
        _("preparation time"),
        default=timedelta(0),
        help_text=_("Time to prepare before the event. (HH:MM:SS)"),
    )
    recovery_time = models.DurationField(
        _("recovery time"),
        default=timedelta(0),
        help_text=_("Time to recover after the event. (HH:MM:SS)"),
    )
    blocks_all_resources = models.BooleanField(
        _("blocks all resources"),
        default=False,
        help_text=_("If checked, the event will block all resources, not just the selected ones."),
    )
    resources = models.ManyToManyField(
        Resource, blank=True, related_name="calendar_events", verbose_name=_("resources")
    )
    resource_groups = models.ManyToManyField(
        ResourceGroup, blank=True, related_name="calendar_events", verbose_name=_("resource groups")
    )
    is_canceled = models.BooleanField(_("is canceled"), default=False)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("calendar event")
        verbose_name_plural = _("calendar events")
        ordering = ("start_date", "start_time")

    def __str__(self) -> str:
        return "{start} - {end} {name}".format(
            start=(
                date_format(self.start, "SHORT_DATETIME_FORMAT")
                if self.start_time
                else date_format(self.start_date, "SHORT_DATE_FORMAT")
            ),
            end=(
                (
                    date_format(self.end, "SHORT_DATETIME_FORMAT")
                    if self.end_date != self.start_date
                    else time_format(self.end_time)
                )
                if self.end_time
                else date_format(self.end_date, "SHORT_DATE_FORMAT")
            ),
            name=self.name,
        )

    def clean(self):
        errors = {}
        if self.start_date is not None and self.end_date is not None:
            if self.start_date > self.end_date:
                errors.setdefault("start_date", []).append(_("Start date must be before end date."))
                errors.setdefault("end_date", []).append(_("End date must be after start date."))
            if (self.start_time is None) ^ (self.end_time is None):
                if self.start_time is None:
                    errors.setdefault("start_time", []).append(_("Start time must be set for timed events."))
                    errors.setdefault("end_time", []).append(_("End time must be empty for all day events."))
                else:
                    errors.setdefault("start_time", []).append(_("Start time must be empty for all day events."))
                    errors.setdefault("end_time", []).append(_("End time must be set for timed events."))
            if self.start_date == self.end_date and self.start_time is not None and self.start_time > self.end_time:
                errors.setdefault("start_time", []).append(_("Start time must be before end time."))
                errors.setdefault("end_time", []).append(_("End time must be after start time."))
        if errors:
            raise ValidationError(errors)

    def save(self, *args, **kwargs):
        self.effective_start = self.start - self.preparation_time if self.start_time else self.start
        self.effective_end = self.end + self.recovery_time if self.end_time else self.end
        super().save(*args, **kwargs)

    @property
    def start(self) -> datetime:
        return datetime.combine(self.start_date, time(0) if self.start_time is None else self.start_time)

    @property
    def end(self) -> datetime:
        if self.end_time:
            return datetime.combine(self.end_date, self.end_time)
        return datetime.combine(self.end_date + timedelta(days=1), time(0))

    @property
    def timeslot(self) -> TimeSlot:
        return TimeSlot(self.effective_start, self.effective_end)

    @cached_property
    def simple_event(self) -> SimpleEvent:
        return SimpleEvent(
            timeslot=self.timeslot,
            resource_groups=[{resource.id} for resource in self.resources.all()]
            + [
                {resource.id for resource in resource_group.resources.all()}
                for resource_group in self.resource_groups.all()
            ],
        )

    @cached_property
    def all_resources(self) -> list[Resource]:
        return list(self.resources.all())

    @cached_property
    def all_resource_groups(self) -> list[ResourceGroup]:
        return list(self.resource_groups.all())

    @cached_property
    def url(self) -> str:
        leprikon_site = LeprikonSite.objects.get_current()
        uri = reverse("admin:leprikon_calendarevent_change", args=(self.pk,))
        return leprikon_site.url + uri

    @cached_property
    def description(self) -> str:
        template = get_template("leprikon/calendar-export.description.txt")
        return template.render({"event": self})

    def has_conflicting_events(self) -> bool:
        if self.is_canceled:
            return False
        relevant_calendar_events_by_timeslot = CalendarEvent.objects.filter(
            effective_start__lt=self.effective_end,
            effective_end__gt=self.effective_start,
            is_canceled=False,
        )
        if relevant_calendar_events_by_timeslot.filter(blocks_all_resources=True).exists():
            return True

        required_resource_groups = self.simple_event.resource_groups
        relevant_resource_ids: set[int] = set(chain.from_iterable(required_resource_groups))
        relevant_calendar_events = relevant_calendar_events_by_timeslot.filter(blocks_all_resources=False).filter(
            models.Q(resources__in=relevant_resource_ids)
            | models.Q(resource_groups__resources__in=relevant_resource_ids)
        )
        relevant_resources = Resource.objects.filter(id__in=relevant_resource_ids).prefetch_related("availabilities")

        events: list[SimpleEvent] = [event.simple_event for event in relevant_calendar_events]

        # unavailable times for each relevant resource
        start_date = self.effective_start.date()
        end_date = self.effective_end.date()
        for resource in relevant_resources:
            available_timeslots = get_time_slots_by_weekly_times(resource.weekly_times, start_date, end_date)
            events.extend(
                SimpleEvent(time_slot, [{resource.id}])
                for time_slot in get_reverse_time_slots(available_timeslots, start_date, end_date)
            )

        return bool(TimeSlots(get_conflicting_timeslots(events)) & self.timeslot)


class CalendarExport(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(_("name"), max_length=255)
    resources = models.ManyToManyField(
        Resource,
        blank=True,
        related_name="calendar_exports",
        verbose_name=_("resources"),
        help_text=_("Select resources to export calendar events for. Leave empty to export events for all resources."),
    )
    export_past_days = models.PositiveSmallIntegerField(_("export past days"), default=30)
    limit_events_count = models.PositiveSmallIntegerField(
        _("limit events count"),
        default=1000,
        help_text=_("Limit the number of events to export. Huge exports may fail to import to Your calendar."),
    )

    class Meta:
        app_label = "leprikon"
        verbose_name = _("calendar export")
        verbose_name_plural = _("calendar exports")

    def __str__(self):
        return self.name

    @cached_property
    def resource_ids(self) -> set[int]:
        return set(self.resources.values_list("id", flat=True))

    @property
    def relevant_events(self) -> models.QuerySet[CalendarEvent]:
        qs = CalendarEvent.objects.filter(
            end_date__gte=date.today() - timedelta(days=self.export_past_days),
            is_canceled=False,
        )
        if self.resource_ids:
            qs = qs.filter(
                models.Q(blocks_all_resources=True)
                | models.Q(resources__in=self.resource_ids)
                | models.Q(resource_groups__resources__in=self.resource_ids)
            )
        return qs

    def get_ical(self) -> str:
        calendar = Calendar()
        calendar.add("prodid", "-//Leprikon//Calendar Export//EN")
        calendar.add("version", "2.0")
        calendar.add("calscale", "GREGORIAN")
        calendar.add("method", "PUBLISH")
        calendar.add("x-wr-calname", self.name)
        calendar.add("x-wr-timezone", settings.TIME_ZONE)
        for event in self.relevant_events[: self.limit_events_count]:
            ical_event = Event()
            ical_event.add("summary", event.name)
            ical_event.add("dtstart", event.start if event.start_time else event.start_date)
            # end date is exclusive if end time is not set
            ical_event.add("dtend", event.end if event.end_time else event.end_date + timedelta(days=1))
            ical_event.add("description", event.description)
            ical_event.add("url", event.url)
            calendar.add_component(ical_event)
        return calendar.to_ical().decode("utf-8")



from collections import namedtuple
from json import loads

from cms.models import CMSPlugin
from cms.models.fields import PageField
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import NoReverseMatch, reverse
from django.utils.formats import date_format
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from djangocms_text_ckeditor.fields import HTMLField
from filer.fields.image import FilerImageField

from leprikon.models.targetgroup import TargetGroup

from ..conf import settings
from ..forms.leaders import LeaderFilterForm
from ..utils import attributes
from .agegroup import AgeGroup
from .citizenship import Citizenship
from .fields import BirthNumberField, EmailField, PostalCodeField
from .school import School
from .schoolyear import SchoolYear


class Leader(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT, related_name="leprikon_leader", verbose_name=_("user")
    )
    description = HTMLField(_("description"), blank=True, default="")
    photo = FilerImageField(verbose_name=_("photo"), blank=True, null=True, related_name="+", on_delete=models.SET_NULL)
    page = PageField(verbose_name=_("page"), blank=True, null=True, related_name="+", on_delete=models.SET_NULL)
    school_years = models.ManyToManyField(SchoolYear, related_name="leaders", verbose_name=_("school years"))

    class Meta:
        app_label = "leprikon"
        ordering = ("user__last_name", "user__first_name")
        verbose_name = _("leader")
        verbose_name_plural = _("leaders")

    def __str__(self):
        return self.full_name or self.user.username

    @cached_property
    def first_name(self):
        return self.user.first_name

    @cached_property
    def last_name(self):
        return self.user.last_name

    @cached_property
    def full_name(self):
        return "{} {}".format(self.first_name, self.last_name).strip()

    @cached_property
    def all_contacts(self):
        return list(self.contacts.all())

    @cached_property
    def all_public_contacts(self):
        return list(self.contacts.filter(public=True))

    @cached_property
    def all_school_years(self):
        return list(self.school_years.all())

    def get_alternate_leader_entries(self, school_year):
        from .journals import JournalLeaderEntry

        return JournalLeaderEntry.objects.filter(
            timesheet__leader=self,
            journal_entry__journal__activity__school_year=school_year,
        ).exclude(journal_entry__journal__activity__in=self.activities.all())

    ActivitiesGroup = namedtuple("ActivitiesGroup", ("activity_type", "activities"))

    def get_activities_by_types(self):
        from .activities import ActivityType

        return (
            self.ActivitiesGroup(
                activity_type=activity_type, activities=activity_type.activities.filter(public=True, leaders=self)
            )
            for activity_type in ActivityType.objects.all()
        )

    def get_absolute_url(self):
        try:
            return reverse("leprikon_leaders:leader_detail", kwargs={"slug": self.user.username})
        except NoReverseMatch:
            return ""


class Contact(models.Model):
    leader = models.ForeignKey(Leader, on_delete=models.CASCADE, related_name="contacts", verbose_name=_("leader"))
    contact_type = models.CharField(_("contact type"), max_length=30, choices=settings.LEPRIKON_CONTACT_TYPES)
    contact = models.CharField(_("contact"), max_length=250)
    order = models.IntegerField(_("order"), blank=True, default=0)
    public = models.BooleanField(_("public"), default=False)

    CONTACT_TYPES = dict(settings.LEPRIKON_CONTACT_TYPES)

    class Meta:
        app_label = "leprikon"
        ordering = ("order",)
        verbose_name = _("contact")
        verbose_name_plural = _("contacts")

    def __str__(self):
        return "{}, {}: {}".format(self.leader.full_name, self.contact_type_name, self.contact)

    @cached_property
    def contact_type_name(self):
        return self.CONTACT_TYPES[self.contact_type]


class Parent(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="leprikon_parents", verbose_name=_("user")
    )
    first_name = models.CharField(_("first name"), max_length=30)
    last_name = models.CharField(_("last name"), max_length=30)
    street = models.CharField(_("street"), max_length=150)
    city = models.CharField(_("city"), max_length=150)
    postal_code = PostalCodeField(_("postal code"))
    email = EmailField(_("email address"), blank=True, default="")
    phone = models.CharField(_("phone"), max_length=30)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("parent")
        verbose_name_plural = _("parents")

    def __str__(self):
        return self.full_name

    @cached_property
    def address(self):
        return "{}, {}, {}".format(self.street, self.city, self.postal_code)

    @cached_property
    def contact(self):
        if self.email and self.phone:
            return "{}, {}".format(self.phone, self.email)
        else:
            return self.email or self.phone or ""

    @cached_property
    def full_name(self):
        return "{} {}".format(self.first_name, self.last_name)


class Participant(models.Model):
    MALE = "m"
    FEMALE = "f"
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="leprikon_participants", verbose_name=_("user")
    )
    age_group = models.ForeignKey(AgeGroup, on_delete=models.PROTECT, related_name="+", verbose_name=_("age group"))
    first_name = models.CharField(_("first name"), max_length=30)
    last_name = models.CharField(_("last name"), max_length=30)
    citizenship = models.ForeignKey(
        Citizenship, on_delete=models.PROTECT, related_name="+", verbose_name=_("citizenship")
    )
    birth_num = BirthNumberField(_("birth number"), blank=True, null=True)
    birth_date = models.DateField(_("birth date"))
    gender = models.CharField(
        _("gender"), max_length=1, choices=((MALE, _("male / boy")), (FEMALE, _("female / girl")))
    )
    street = models.CharField(_("street"), max_length=150)
    city = models.CharField(_("city"), max_length=150)
    postal_code = PostalCodeField(_("postal code"))
    email = EmailField(_("email address"), blank=True, default="")
    phone = models.CharField(_("phone"), max_length=30, blank=True, default="")
    school = models.ForeignKey(
        School, blank=True, null=True, on_delete=models.PROTECT, related_name="participants", verbose_name=_("school")
    )
    school_other = models.CharField(_("other school"), max_length=150, blank=True, default="")
    school_class = models.CharField(_("class"), max_length=30, blank=True, default="")
    health = models.TextField(_("health"), blank=True, default="")
    answers = models.TextField(_("additional answers"), blank=True, default="{}", editable=False)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("participant")
        verbose_name_plural = _("participants")

    def __str__(self):
        return _("{first_name} {last_name} ({birth_date})").format(
            first_name=self.first_name,
            last_name=self.last_name,
            birth_date=date_format(self.birth_date, "SHORT_DATE_FORMAT"),
        )

    def get_answers(self):
        return loads(self.answers)

    @cached_property
    def full_name(self):
        return "{} {}".format(self.first_name, self.last_name)

    @cached_property
    def address(self):
        return "{}, {}, {}".format(self.street, self.city, self.postal_code)

    @cached_property
    def contact(self) -> str:
        if self.email and self.phone:
            return "{}, {}".format(self.phone, self.email)
        else:
            return self.email or self.phone or ""

    @cached_property
    def school_name(self) -> str:
        return self.school and str(self.school) or self.school_other

    @cached_property
    def school_and_class(self) -> str:
        if self.school_name and self.school_class:
            return "{}, {}".format(self.school_name, self.school_class)
        else:
            return self.school_name or self.school_class or ""


class GroupContact(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="leprikon_group_contacts",
        verbose_name=_("user"),
    )
    target_group = models.ForeignKey(
        TargetGroup, on_delete=models.PROTECT, related_name="+", verbose_name=_("target group")
    )
    name = models.CharField(_("group name"), blank=True, default="", max_length=150)
    first_name = models.CharField(_("first name"), max_length=30)
    last_name = models.CharField(_("last name"), max_length=30)
    street = models.CharField(_("street"), max_length=150)
    city = models.CharField(_("city"), max_length=150)
    postal_code = PostalCodeField(_("postal code"))
    phone = models.CharField(_("phone"), max_length=30)
    email = EmailField(_("email address"))
    school = models.ForeignKey(
        School, blank=True, null=True, on_delete=models.PROTECT, related_name="+", verbose_name=_("school")
    )
    school_other = models.CharField(_("other school"), max_length=150, blank=True, default="")
    school_class = models.CharField(_("class"), max_length=30, blank=True, default="")
    answers = models.TextField(_("additional answers"), blank=True, default="{}", editable=False)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("Group contact")
        verbose_name_plural = _("Group contacts")

    def __str__(self):
        return f"{self.name} ({self.full_name})" if self.name else self.full_name

    def get_answers(self):
        return loads(self.answers)

    @cached_property
    def full_name(self):
        return "{} {}".format(self.first_name, self.last_name)

    @cached_property
    def address(self):
        return "{}, {}, {}".format(self.street, self.city, self.postal_code)

    @cached_property
    def contact(self) -> str:
        if self.email and self.phone:
            return "{}, {}".format(self.phone, self.email)
        else:
            return self.email or self.phone or ""

    @cached_property
    def school_name(self) -> str:
        return self.school and str(self.school) or self.school_other

    @cached_property
    def school_and_class(self) -> str:
        if self.school_name and self.school_class:
            return "{}, {}".format(self.school_name, self.school_class)
        else:
            return self.school_name or self.school_class or ""


class BillingInfo(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="leprikon_billing_info", verbose_name=_("user")
    )
    name = models.CharField(_("name"), max_length=150)
    street = models.CharField(_("street"), max_length=150, blank=True, default="")
    city = models.CharField(_("city"), max_length=150, blank=True, default="")
    postal_code = PostalCodeField(_("postal code"), blank=True, default="")
    company_num = models.CharField(_("company number"), max_length=8, blank=True, default="")
    vat_number = models.CharField(_("VAT number"), max_length=12, blank=True, default="")
    contact_person = models.CharField(_("contact person"), max_length=60, blank=True, default="")
    phone = models.CharField(_("phone"), max_length=30, blank=True, default="")
    email = EmailField(_("email address"), blank=True, default="")
    employee = models.CharField(_("employee ID"), max_length=150, blank=True, default="")

    class Meta:
        app_label = "leprikon"
        ordering = ("name",)
        verbose_name = _("billing information")
        verbose_name_plural = _("billing information")

    def __str__(self):
        return self.name

    @attributes(short_description=_("address"))
    @cached_property
    def address(self):
        return ", ".join(filter(bool, (self.street, self.city, self.postal_code)))


class LeaderPlugin(CMSPlugin):
    leader = models.ForeignKey(Leader, on_delete=models.CASCADE, related_name="+", verbose_name=_("leader"))
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_LEADER_TEMPLATES,
        default=settings.LEPRIKON_LEADER_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"


class LeaderListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    activity = models.ForeignKey(
        "leprikon.Activity",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("activity"),
    )
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_LEADERLIST_TEMPLATES,
        default=settings.LEPRIKON_LEADERLIST_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"

    def clean(self):
        if self.school_year and self.activity and self.activity.school_year != self.school_year:
            raise ValidationError(
                {
                    "school_year": [_("Selected activity is not in the selected school year.")],
                    "activity": [_("Selected activity is not in the selected school year.")],
                }
            )

    def render(self, context):
        if self.activity:
            leaders = self.activity.leaders.all()
        else:
            school_year = (
                self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
            )
            leaders = school_year.leaders.all()
        context.update(
            {
                "leaders": leaders,
            }
        )
        return context


class FilteredLeaderListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )

    class Meta:
        app_label = "leprikon"

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )
        form = LeaderFilterForm(
            school_year=school_year,
            data=context["request"].GET,
        )
        context.update(
            {
                "school_year": school_year,
                "form": form,
                "leaders": form.get_queryset(),
            }
        )
        return context



from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _


class School(models.Model):
    name = models.CharField(_("name"), max_length=50)
    street = models.CharField(_("street"), max_length=50, blank=True, default="")
    city = models.CharField(_("city"), max_length=50, blank=True, default="")

    class Meta:
        app_label = "leprikon"
        ordering = ("city", "name")
        verbose_name = _("school")
        verbose_name_plural = _("schools")

    def __str__(self):
        if self.address:
            return "{}, {}".format(self.name, self.address)
        else:
            return self.name

    @cached_property
    def address(self):
        if self.street and self.city:
            return "{}, {}".format(self.street, self.city)
        else:
            return self.street or self.city



from django.db.migrations.operations.models import ModelOperation


class AlterModelBases(ModelOperation):
    reduce_to_sql = False
    reversible = True

    def __init__(self, name, bases):
        self.bases = bases
        super().__init__(name)

    def state_forwards(self, app_label, state):
        """
        Overwrite a models base classes with a custom list of
        bases instead, then force Django to reload the model
        with this (probably completely) different class hierarchy.
        """
        state.models[app_label, self.name_lower].bases = self.bases
        state.reload_model(app_label, self.name_lower)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        pass

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        pass

    def describe(self):
        return "Update %s bases to %s" % (self.name, self.bases)



from django.utils.formats import date_format
from django.utils.safestring import mark_safe
from django.utils.timezone import localtime

from ..utils import attributes


def datetime_with_by(attr, short_description):
    @attributes(__name__=f"{attr}_with_by", admin_order_field=attr, short_description=short_description)
    def f(self, obj):
        d = getattr(obj, attr)
        if d:
            d_formated = date_format(localtime(d), "SHORT_DATETIME_FORMAT")
            by = getattr(obj, f"{attr}_by")
            return mark_safe(f'<span title="{by}">{d_formated}</span>') if by else d_formated
        return "-"

    return f



from django.db import migrations, models
import django.db.models.deletion


def set_journalentry_period(apps, schema_editor):
    SchoolYearDivision = apps.get_model("leprikon", "SchoolYearDivision")
    SchoolYearPeriod = apps.get_model("leprikon", "SchoolYearPeriod")
    JournalEntry = apps.get_model("leprikon", "JournalEntry")
    for school_year_period in SchoolYearPeriod.objects.filter(start__isnull=False, end__isnull=False):
        JournalEntry.objects.filter(
            journal__school_year_division=school_year_period.school_year_division,
            date__gte=school_year_period.start,
            date__lte=school_year_period.end,
        ).update(period=school_year_period)
    for school_year_division in SchoolYearDivision.objects.all():
        JournalEntry.objects.filter(
            journal__school_year_division=school_year_division,
            period__isnull=True,
        ).update(period=SchoolYearPeriod.objects.filter(school_year_division=school_year_division).first())


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0075_journal_school_year_division"),
    ]

    operations = [
        migrations.AddField(
            model_name="journalentry",
            name="period",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="journal_entries",
                to="leprikon.schoolyearperiod",
                verbose_name="school year period",
            ),
        ),
        migrations.RunPython(set_journalentry_period),
    ]



from collections import defaultdict, namedtuple
from datetime import timedelta
from functools import lru_cache

from django.db.models import Sum
from django.template.response import TemplateResponse
from django.urls import reverse_lazy as reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ...forms.reports.courses import CoursePaymentsForm, CoursePaymentsStatusForm, CourseStatsForm
from ...models.activities import (
    ActivityModel,
    ActivityTime,
    Payment,
    RegistrationParticipant,
)
from ...models.citizenship import Citizenship
from ...models.courses import Course, CourseRegistration
from ...models.journals import JournalTime
from ...models.roles import Participant
from ...models.statgroup import StatGroup
from ...views.generic import FormView


class ReportCoursePaymentsView(FormView):
    form_class = CoursePaymentsForm
    template_name = "leprikon/reports/course_payments.html"
    title = _("Course payments")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    def form_valid(self, form):
        context = form.cleaned_data
        context["form"] = form
        context["received_payments"] = Payment.objects.filter(
            target_registration__activity__activity_type__model=ActivityModel.COURSE,
            accounted__gte=context["date_start"],
            accounted__lte=context["date_end"],
        )
        context["returned_payments"] = Payment.objects.filter(
            source_registration__activity__activity_type__model=ActivityModel.COURSE,
            accounted__gte=context["date_start"],
            accounted__lte=context["date_end"],
        )
        context["received_payments_sum"] = context["received_payments"].aggregate(sum=Sum("amount"))["sum"] or 0
        context["returned_payments_sum"] = context["returned_payments"].aggregate(sum=Sum("amount"))["sum"] or 0
        context["sum"] = context["received_payments_sum"] - context["returned_payments_sum"]
        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))


class ReportCoursePaymentsStatusView(FormView):
    form_class = CoursePaymentsStatusForm
    template_name = "leprikon/reports/course_payments_status.html"
    title = _("Course payments status")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    CoursePaymentsStatusSums = namedtuple("CoursePaymentsStatusSums", ("registrations", "status"))

    def form_valid(self, form):
        context = form.cleaned_data
        context["form"] = form
        context["reports"] = [
            self.Report(course, context["date"])
            for course in Course.objects.filter(school_year=self.request.school_year)
        ]
        context["sum"] = self.CoursePaymentsStatusSums(
            registrations=sum(len(r.registration_statuses) for r in context["reports"]),
            status=sum(r.status for r in context["reports"]),
        )
        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))

    class Report:
        def __init__(self, course, d):
            self.course = course
            self.date = d

        RegPaymentStatus = namedtuple("RegPaymentStatus", ("registration", "status"))

        @cached_property
        def registration_statuses(self):
            return [
                registration_status
                for registration_status in (
                    self.RegPaymentStatus(
                        registration=registration,
                        status=registration.get_payment_status(self.date),
                    )
                    for registration in CourseRegistration.objects.filter(
                        activity=self.course,
                        approved__date__lte=self.date,
                    )
                )
                if registration_status.status.receivable
            ]

        @cached_property
        def status(self):
            return sum(rs.status for rs in self.registration_statuses)


class ReportCourseStatsView(FormView):
    form_class = CourseStatsForm
    template_name = "leprikon/reports/course_stats.html"
    title = _("Course statistics")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    ReportItem = namedtuple("ReportItem", ("stat_group", "all", "boys", "girls", "citizenships"))

    _journal_deltas = {}
    _activity_deltas = {}

    def get_delta(self, cache, model, **kwargs):
        [id] = kwargs.values()
        try:
            return cache[id]
        except KeyError:
            cache[id] = sum(
                (t.delta for t in model.objects.filter(**kwargs)),
                start=timedelta(0),
            )
        return cache[id]

    @lru_cache
    def get_journal_delta(self, journal_id):
        return self.get_delta(self._journal_deltas, JournalTime, journal_id=journal_id)

    @lru_cache
    def get_activity_delta(self, activity_id):
        return self.get_delta(self._activity_deltas, ActivityTime, activity_id=activity_id)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["school_year"] = self.request.school_year
        return kwargs

    def form_valid(self, form):
        d = form.cleaned_data["date"]
        paid_only = form.cleaned_data["paid_only"]
        paid_later = form.cleaned_data["paid_later"]
        approved_later = form.cleaned_data["approved_later"]
        unique_participants = form.cleaned_data["unique_participants"]
        max_weekly_hours = form.cleaned_data["max_weekly_hours"]
        context = form.cleaned_data
        context["form"] = form

        if approved_later:
            # approved registrations created by the date
            participants = RegistrationParticipant.objects.filter(
                registration__created__date__lte=d,
                registration__approved__isnull=False,
            )
        else:
            # registrations approved by the date
            participants = RegistrationParticipant.objects.filter(
                registration__approved__date__lte=d,
            )
        participants = (
            participants.filter(
                registration__activity__in=form.cleaned_data["courses"],
            )
            .exclude(registration__canceled__date__lte=d)
            .select_related("registration", "age_group")
        )
        if paid_only:
            paid_date = None if paid_later else d
            participants = [
                participant
                for participant in participants
                if participant.registration.courseregistration.get_payment_status(paid_date).amount_due == 0
            ]
        else:
            participants = list(participants)

        context["courses_count"] = len(set(participant.registration.activity_id for participant in participants))

        weekly_delta_by_participant = defaultdict(timedelta)
        for participant in participants:
            weekly_delta_by_participant[participant.key] += sum(
                (self.get_journal_delta(journal.id) for journal in participant.journals.all()),
                start=timedelta(0),
            ) or self.get_activity_delta(participant.registration.activity_id)

        delta = sum(
            weekly_delta_by_participant.values(),
            start=timedelta(0),
        )
        context["participant_hours_count"] = delta.days * 24 + delta.seconds / 3600

        if unique_participants or max_weekly_hours:
            participants_by_key = {p.key: p for p in participants}
            if max_weekly_hours:
                max_weekly_delta = timedelta(hours=max_weekly_hours)
                participants = list(
                    p for p in participants_by_key.values() if weekly_delta_by_participant[p.key] <= max_weekly_delta
                )
            else:
                participants = list(participants_by_key.values())

        citizenships = list(Citizenship.objects.all())
        context["citizenships"] = citizenships

        context["participants_counts"] = self.ReportItem(
            stat_group=None,
            all=len(participants),
            boys=len([p for p in participants if p.gender == Participant.MALE]),
            girls=len([p for p in participants if p.gender == Participant.FEMALE]),
            citizenships=[
                len([p for p in participants if p.citizenship_id == citizenship.id]) for citizenship in citizenships
            ],
        )
        context["participants_counts_by_stat_groups"] = []
        for stat_group in StatGroup.objects.all():
            parts = [p for p in participants if p.age_group.stat_group_id == stat_group.id]
            context["participants_counts_by_stat_groups"].append(
                self.ReportItem(
                    stat_group=stat_group,
                    all=len(parts),
                    boys=len([p for p in parts if p.gender == Participant.MALE]),
                    girls=len([p for p in parts if p.gender == Participant.FEMALE]),
                    citizenships=[
                        len([p for p in parts if p.citizenship_id == citizenship.id]) for citizenship in citizenships
                    ],
                )
            )

        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))



# Generated by Django 1.10.7 on 2017-09-07 16:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0003_managers'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='subjectregistration',
            unique_together=set([]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-07-12 08:05
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
        ("leprikon", "0054_registration_dates_by_cleanup"),
    ]

    operations = [
        migrations.CreateModel(
            name="SchoolYearBlockPlugin",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="leprikon_schoolyearblockplugin",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "school_years",
                    models.ManyToManyField(
                        help_text="Which school years should the block be visible for?",
                        related_name="+",
                        to="leprikon.SchoolYear",
                        verbose_name="school years",
                    ),
                ),
            ],
            bases=("cms.cmsplugin",),
        ),
    ]



# Generated by Django 3.2.25 on 2025-08-28 08:38

from datetime import datetime, time, timedelta
from django.db import migrations, models
import django.utils.timezone


def set_effective_times(apps, schema_editor):
    CalendarEvent = apps.get_model("leprikon", "CalendarEvent")
    for event in CalendarEvent.objects.all():
        if event.start_time is None or event.end_time is None:
            event.start_time = None
            event.end_time = None
            event.effective_start = django.utils.timezone.make_aware(
                datetime.combine(event.start_date, time(0)),
            )
            event.effective_end = django.utils.timezone.make_aware(
                datetime.combine(event.end_date, time(0)) + timedelta(days=1),
            )
        else:
            event.effective_start = django.utils.timezone.make_aware(
                datetime.combine(event.start_date, event.start_time) - event.preparation_time,
            )
            event.effective_end = django.utils.timezone.make_aware(
                datetime.combine(event.end_date, event.end_time) + event.recovery_time,
            )
        event.save()


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0091_calendar_updates"),
    ]

    operations = [
        migrations.AddField(
            model_name="calendarevent",
            name="effective_end",
            field=models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name="effective end"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="calendarevent",
            name="effective_start",
            field=models.DateTimeField(
                default=django.utils.timezone.now, editable=False, verbose_name="effective start"
            ),
            preserve_default=False,
        ),
        migrations.RunPython(set_effective_times, reverse_code=migrations.RunPython.noop),
    ]



from adminsortable2.admin import SortableAdminMixin
from django.contrib import admin

from ..models.citizenship import Citizenship


@admin.register(Citizenship)
class CitizenshipAdmin(SortableAdminMixin, admin.ModelAdmin):
    list_display = ("name", "require_birth_num")
    list_editable = ("require_birth_num",)



# Generated by Django 3.2.25 on 2025-07-28 17:43

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0090_alter_registration_calendar_event"),
    ]

    operations = [
        migrations.AddField(
            model_name="calendarevent",
            name="blocks_all_resources",
            field=models.BooleanField(
                default=False,
                help_text="If checked, the event will block all resources, not just the selected ones.",
                verbose_name="blocks all resources",
            ),
        ),
        migrations.CreateModel(
            name="CalendarExport",
            fields=[
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=255, verbose_name="name")),
                ("export_past_days", models.PositiveSmallIntegerField(default=30, verbose_name="export past days")),
                (
                    "limit_events_count",
                    models.PositiveSmallIntegerField(
                        default=1000,
                        help_text="Limit the number of events to export. Huge exports may fail to import to Your calendar.",
                        verbose_name="limit events count",
                    ),
                ),
                (
                    "resources",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Select resources to export calendar events for. Leave empty to export events for all resources.",
                        related_name="calendar_exports",
                        to="leprikon.Resource",
                        verbose_name="resources",
                    ),
                ),
            ],
            options={
                "verbose_name": "calendar export",
                "verbose_name_plural": "calendar exports",
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.27 on 2020-02-16 22:13
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import leprikon.models.fields
import localflavor.generic.models


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0046_attachment_events'),
    ]

    operations = [
        migrations.AddField(
            model_name='leprikonsite',
            name='pr_print_setup',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='leprikon.PrintSetup', verbose_name='default payment request print setup'),
        ),
        migrations.AddField(
            model_name='subject',
            name='pr_print_setup',
            field=models.ForeignKey(blank=True, help_text='Only use to set value specific for this subject.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='leprikon.PrintSetup', verbose_name='payment request print setup'),
        ),
        migrations.AddField(
            model_name='subjecttype',
            name='pr_print_setup',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='leprikon.PrintSetup', verbose_name='payment request print setup'),
        ),
        migrations.AlterField(
            model_name='billinginfo',
            name='city',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='city'),
        ),
        migrations.AlterField(
            model_name='billinginfo',
            name='company_num',
            field=models.CharField(blank=True, default='', max_length=8, verbose_name='company number'),
        ),
        migrations.AlterField(
            model_name='billinginfo',
            name='contact_person',
            field=models.CharField(blank=True, default='', max_length=60, verbose_name='contact person'),
        ),
        migrations.AlterField(
            model_name='billinginfo',
            name='postal_code',
            field=leprikon.models.fields.PostalCodeField(blank=True, default='', verbose_name='postal code'),
        ),
        migrations.AlterField(
            model_name='billinginfo',
            name='street',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='street'),
        ),
        migrations.AlterField(
            model_name='billinginfo',
            name='vat_number',
            field=models.CharField(blank=True, default='', max_length=12, verbose_name='VAT number'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='bic',
            field=localflavor.generic.models.BICField(blank=True, default='', max_length=11, verbose_name='BIC (SWIFT)'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='city',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='city'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='company_num',
            field=models.CharField(blank=True, default='', max_length=8, verbose_name='company number'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='email',
            field=leprikon.models.fields.EmailField(blank=True, default='', max_length=254, verbose_name='email address'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='iban',
            field=localflavor.generic.models.IBANField(blank=True, default='', include_countries=None, max_length=34, use_nordea_extensions=False, verbose_name='IBAN'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='phone',
            field=models.CharField(blank=True, default='', max_length=30, verbose_name='phone'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='postal_code',
            field=leprikon.models.fields.PostalCodeField(blank=True, default='', verbose_name='postal code'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='street',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='street'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='vat_number',
            field=models.CharField(blank=True, default='', max_length=12, verbose_name='VAT number'),
        ),
        migrations.AlterField(
            model_name='subjectregistrationbillinginfo',
            name='city',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='city'),
        ),
        migrations.AlterField(
            model_name='subjectregistrationbillinginfo',
            name='company_num',
            field=models.CharField(blank=True, default='', max_length=8, verbose_name='company number'),
        ),
        migrations.AlterField(
            model_name='subjectregistrationbillinginfo',
            name='contact_person',
            field=models.CharField(blank=True, default='', max_length=60, verbose_name='contact person'),
        ),
        migrations.AlterField(
            model_name='subjectregistrationbillinginfo',
            name='postal_code',
            field=leprikon.models.fields.PostalCodeField(blank=True, default='', verbose_name='postal code'),
        ),
        migrations.AlterField(
            model_name='subjectregistrationbillinginfo',
            name='street',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='street'),
        ),
        migrations.AlterField(
            model_name='subjectregistrationbillinginfo',
            name='vat_number',
            field=models.CharField(blank=True, default='', max_length=12, verbose_name='VAT number'),
        ),
    ]



from django import forms


class RegistrationLinkAdminForm(forms.ModelForm):
    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)

        # limit choices of activities
        activities_choices = self.fields["activity_variants"].widget.choices
        activities_choices.queryset = activities_choices.queryset.filter(
            activity__school_year=self.school_year,
            activity__activity_type=self.activity_type,
        )



from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from localflavor.generic.models import BICField, IBANField

from ..utils import attributes
from .fields import EmailField, PostalCodeField
from .printsetup import PrintSetup
from .utils import BankAccount


class Organization(models.Model):
    name = models.CharField(_("organization name"), max_length=150)
    street = models.CharField(_("street"), max_length=150, blank=True, default="")
    city = models.CharField(_("city"), max_length=150, blank=True, default="")
    postal_code = PostalCodeField(_("postal code"), blank=True, default="")
    email = EmailField(_("email address"), blank=True, default="")
    phone = models.CharField(_("phone"), max_length=30, blank=True, default="")
    company_num = models.CharField(_("company number"), max_length=8, blank=True, default="")
    vat_number = models.CharField(_("VAT number"), max_length=12, blank=True, default="")
    iban = IBANField(_("IBAN"), blank=True, default="")
    bic = BICField(_("BIC (SWIFT)"), blank=True, default="")
    donation_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("donation print setup"),
    )

    class Meta:
        app_label = "leprikon"
        verbose_name = _("organization")
        verbose_name_plural = _("organizations")

    def __str__(self):
        return f"{self.name} ({self.bank_account})"

    @attributes(short_description=_("address"))
    @cached_property
    def address(self):
        return ", ".join(filter(bool, (self.street, self.city, self.postal_code)))

    @attributes(short_description=_("bank account"))
    @cached_property
    def bank_account(self):
        return self.iban and BankAccount(self.iban)



from typing import Any, Dict
from urllib.parse import urlencode

from bankreader.models import Transaction as BankreaderTransaction
from django.contrib import admin
from django.contrib.auth.decorators import permission_required
from django.http import HttpRequest, HttpResponse, HttpResponseBadRequest, HttpResponseRedirect, JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import path, reverse
from django.utils import timezone
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ..models.transaction import Transaction
from ..utils import amount_color, attributes, currency
from .export import AdminExportMixin
from .pdf import PdfExportAdminMixin
from .sendmail import SendMailAdminMixin
from .utils import datetime_with_by


class TransactionTypeListFilter(admin.ChoicesFieldListFilter):
    def choices(self, changelist):
        yield {
            "selected": self.lookup_val is None,
            "query_string": changelist.get_query_string({}, [self.lookup_kwarg, self.lookup_kwarg_isnull]),
            "display": _("All"),
        }
        none_title = ""
        for lookup, title in self.field.flatchoices:
            if lookup is None:
                none_title = title
                continue
            if self.field.model.transaction_types and lookup not in self.field.model.transaction_types:
                continue
            yield {
                "selected": str(lookup) == self.lookup_val,
                "query_string": changelist.get_query_string({self.lookup_kwarg: lookup}, [self.lookup_kwarg_isnull]),
                "display": title,
            }
        if none_title:
            yield {
                "selected": bool(self.lookup_val_isnull),
                "query_string": changelist.get_query_string(
                    {
                        self.lookup_kwarg_isnull: "True",
                    },
                    [self.lookup_kwarg],
                ),
                "display": none_title,
            }


class TransactionAdminMixin:
    date_hierarchy = "accounted"
    ordering = ("-accounted",)

    accounted_with_by = datetime_with_by("accounted", _("accounted time"))
    last_updated_with_by = datetime_with_by("last_updated", _("last updated time"))

    def is_closed(self, request, obj):
        return (
            obj
            and request.leprikon_site.max_closure_date
            and request.leprikon_site.max_closure_date > obj.accounted.date()
        )

    def has_delete_permission(self, request, obj=None):
        if self.is_closed(request, obj):
            return False
        else:
            return super().has_delete_permission(request, obj)

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = super().get_readonly_fields(request, obj)
        if obj:
            # it is strange, but obj given to this method contains values from request.POST
            # but we need to decide according to current state in database
            obj = self.model.objects.get(pk=obj.pk)
        if self.is_closed(request, obj):
            return tuple(set(readonly_fields).union(set(self.closed_fields)))
        else:
            return readonly_fields

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:

            def delete_selected(model_admin, request, queryset):
                if request.leprikon_site.max_closure_date:
                    queryset = queryset.filter(accounted__date__gt=request.leprikon_site.max_closure_date)
                return admin.actions.delete_selected(model_admin, request, queryset)

            actions["delete_selected"] = (delete_selected, *actions["delete_selected"][1:])
        return actions

    @attributes(admin_order_field="amount", short_description=_("amount"))
    def amount_html(self, obj):
        return mark_safe(
            format_html(
                '<b style="color: {color}">{amount}</b>',
                color=amount_color(obj.amount),
                amount=currency(abs(obj.amount)),
            )
        )

    def save_model(self, request, obj, form, change):
        if change:
            obj.last_updated = timezone.now()
            obj.last_updated_by = request.user
        else:
            obj.accounted = timezone.now()
            obj.accounted_by = request.user
        super().save_model(request, obj, form, change)


class TransactionBaseAdmin(
    TransactionAdminMixin, PdfExportAdminMixin, SendMailAdminMixin, AdminExportMixin, admin.ModelAdmin
):
    actions = AdminExportMixin.actions + SendMailAdminMixin.actions + PdfExportAdminMixin.actions
    list_display = ("id", "accounted_with_by", "transaction_type", "amount", "last_updated_with_by", "note")
    closed_fields = ("accounted", "amount", "target_registration", "source_registration", "donor", "organization")
    list_editable = ("note",)
    raw_id_fields = ("target_registration", "source_registration", "bankreader_transaction", "pays_payment", "donor")

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        if self.model.transaction_types:
            form.base_fields["transaction_type"].choices = (
                (value, label)
                for value, label in Transaction.TRANSACTION_TYPE_LABELS.items()
                if value in self.model.transaction_types
            )
        return form

    def get_urls(self):
        populate_view = self.admin_site.admin_view(
            permission_required(f"{self.model._meta.app_label}.add_{self.model._meta.model_name}")(self.populate)
        )
        return [
            path("populate.json", populate_view, name="leprikon_payment_populate"),
        ] + super().get_urls()

    def populate(self, request):
        try:
            return JsonResponse(
                {
                    "amount": get_object_or_404(
                        BankreaderTransaction,
                        id=int(request.GET["bankreader_transaction"]),
                    ).amount
                }
            )
        except (KeyError, ValueError):
            return HttpResponseBadRequest()


@admin.register(Transaction)
class TransactionAdmin(TransactionBaseAdmin):
    def get_model_perms(self, request):
        return {}

    def has_delete_permission(self, request, obj=None):
        return False

    def get_actions(self, request):
        return {}

    def add_view(self, request: HttpRequest, form_url="", extra_context: Dict[str, Any] = None) -> HttpResponse:
        try:
            transaction = get_object_or_404(
                BankreaderTransaction,
                id=int(request.GET["bankreader_transaction"]),
            )
        except (KeyError, ValueError):
            return HttpResponseBadRequest()
        if transaction.amount > 0:
            url = reverse("admin:leprikon_receivedpayment_add")
            transaction_type = Transaction.PAYMENT_BANK
        else:
            url = reverse("admin:leprikon_returnedpayment_add")
            transaction_type = Transaction.RETURN_BANK
        query = urlencode(
            dict(
                bankreader_transaction=transaction.id,
                amount=abs(transaction.amount),
                transaction_type=transaction_type,
            )
        )
        return HttpResponseRedirect(f"{url}?{query}")



from collections import namedtuple
from itertools import chain

from django.template.response import TemplateResponse
from django.urls import reverse_lazy as reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ...forms.reports.debtors import DebtorsForm
from ...models.courses import CourseRegistration
from ...models.events import EventRegistration
from ...models.orderables import OrderableRegistration
from ...views.generic import FormView


class ReportDebtorsView(FormView):
    form_class = DebtorsForm
    template_name = "leprikon/reports/debtors.html"
    title = _("Debtors list")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    ReportItem = namedtuple("ReportItem", ("registration", "status"))

    class Report(list):
        @cached_property
        def sum(self):
            return sum(item.status for item in self)

    def form_valid(self, form):
        context = form.cleaned_data
        context["form"] = form
        context["reports"] = {}
        context["sum"] = 0

        for reg in chain.from_iterable(
            qs.filter(
                activity__school_year=self.request.school_year,
                approved__date__lte=context["date"],
            )
            .prefetch_related(
                "discounts",
                "received_payments",
                "returned_payments",
            )
            .select_related("user")
            for qs in (
                CourseRegistration.objects.prefetch_related("course_registration_periods"),
                EventRegistration.objects,
                OrderableRegistration.objects,
            )
        ):
            status = reg.get_payment_status(context["date"])
            if status.amount_due:
                report = context["reports"].setdefault(reg.user, self.Report())
                report.append(self.ReportItem(registration=reg, status=status))
        context["sum"] = sum(report.sum for report in context["reports"].values())

        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))



# Generated by Django 3.2.16 on 2023-01-20 20:11

from django.db import migrations
import djangocms_text_ckeditor.fields


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0076_journalentry_period'),
    ]

    operations = [
        migrations.AddField(
            model_name='place',
            name='description',
            field=djangocms_text_ckeditor.fields.HTMLField(blank=True, default='', verbose_name='description'),
        ),
    ]



# Generated by Django 1.11.12 on 2018-08-29 21:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0018_citizenship_cleanup"),
    ]

    operations = [
        migrations.CreateModel(
            name="Department",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=50, verbose_name="name")),
                ("code", models.PositiveSmallIntegerField(blank=True, default=0, verbose_name="department code")),
            ],
            options={
                "ordering": ("code", "name"),
                "verbose_name": "department",
                "verbose_name_plural": "departments",
            },
        ),
        migrations.RemoveField(
            model_name="manager",
            name="school_years",
        ),
        migrations.RemoveField(
            model_name="manager",
            name="user",
        ),
        migrations.AlterModelOptions(
            name="course",
            options={"ordering": ("code", "name"), "verbose_name": "course", "verbose_name_plural": "courses"},
        ),
        migrations.AlterModelOptions(
            name="subject",
            options={"ordering": ("code", "name"), "verbose_name": "subject", "verbose_name_plural": "subjects"},
        ),
        migrations.RemoveField(
            model_name="courselistplugin",
            name="managers",
        ),
        migrations.RemoveField(
            model_name="eventlistplugin",
            name="managers",
        ),
        migrations.RemoveField(
            model_name="subject",
            name="managers",
        ),
        migrations.AlterField(
            model_name="subject",
            name="age_groups",
            field=models.ManyToManyField(related_name="subjects", to="leprikon.AgeGroup", verbose_name="age groups"),
        ),
        migrations.DeleteModel(
            name="Manager",
        ),
        migrations.AddField(
            model_name="courselistplugin",
            name="departments",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by departments.",
                related_name="+",
                to="leprikon.Department",
                verbose_name="departments",
            ),
        ),
        migrations.AddField(
            model_name="eventlistplugin",
            name="departments",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by departments.",
                related_name="+",
                to="leprikon.Department",
                verbose_name="departments",
            ),
        ),
        migrations.AddField(
            model_name="subject",
            name="department",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="subjects",
                to="leprikon.Department",
                verbose_name="department",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2021-08-15 10:45
from __future__ import unicode_literals
from leprikon.models.activities import ActivityModel, ActivityType

from django.db import migrations, models
import django.db.models.deletion
import djangocms_text_ckeditor.fields
import leprikon.models.fields
import leprikon.models.startend
import leprikon.models.times


def migrate_journals(apps, schema_editor):
    Journal = apps.get_model("leprikon", "Journal")
    JournalEntry = apps.get_model("leprikon", "JournalEntry")
    Subject = apps.get_model("leprikon", "Subject")
    SubjectRegistrationParticipant = apps.get_model("leprikon", "SubjectRegistrationParticipant")

    for subject in Subject.objects.filter(
        id__in=JournalEntry.objects.values_list("subject_id", flat=True)
    ).select_related("subject_type"):
        journal = Journal.objects.create(
            subject=subject,
            risks=subject.risks,
            plan=subject.plan,
            evaluation=subject.evaluation,
        )
        if subject.subject_type.subject_type == ActivityModel.COURSE:
            journal.participants.set(
                SubjectRegistrationParticipant.objects.filter(
                    registration__approved__isnull=False,
                    registration__courseregistration__course_history__course_id=subject.id,
                )
            )
        else:
            journal.participants.set(
                SubjectRegistrationParticipant.objects.filter(
                    registration__approved__isnull=False,
                    registration__subject=subject,
                )
            )
        journal.leaders.set(subject.leaders.all())
        subject.journal_entries.update(journal=journal)


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0062_refund_offers"),
    ]

    operations = [
        migrations.CreateModel(
            name="Journal",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(blank=True, default="", max_length=150, verbose_name="journal name")),
                ("risks", djangocms_text_ckeditor.fields.HTMLField(blank=True, verbose_name="risks")),
                ("plan", djangocms_text_ckeditor.fields.HTMLField(blank=True, verbose_name="plan")),
                ("evaluation", djangocms_text_ckeditor.fields.HTMLField(blank=True, verbose_name="evaluation")),
                (
                    "participants",
                    models.ManyToManyField(
                        blank=True,
                        related_name="journals",
                        to="leprikon.SubjectRegistrationParticipant",
                        verbose_name="participants",
                    ),
                ),
                (
                    "leaders",
                    models.ManyToManyField(
                        blank=True, related_name="journals", to="leprikon.Leader", verbose_name="leaders"
                    ),
                ),
            ],
            options={
                "verbose_name": "journal",
                "verbose_name_plural": "journal",
            },
            bases=(leprikon.models.times.TimesMixin, models.Model),
        ),
        migrations.CreateModel(
            name="JournalTime",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "day_of_week",
                    models.IntegerField(
                        choices=[
                            (1, "Monday"),
                            (2, "Tuesday"),
                            (3, "Wednesday"),
                            (4, "Thursday"),
                            (5, "Friday"),
                            (6, "Saturday"),
                            (7, "Sunday"),
                        ],
                        verbose_name="day of week",
                    ),
                ),
                ("start", models.TimeField(blank=True, null=True, verbose_name="start time")),
                ("end", models.TimeField(blank=True, null=True, verbose_name="end time")),
                (
                    "journal",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="times",
                        to="leprikon.Journal",
                        verbose_name="journal",
                    ),
                ),
            ],
            options={
                "verbose_name": "time",
                "verbose_name_plural": "times",
                "ordering": ("day_of_week", "start"),
            },
            bases=(leprikon.models.startend.StartEndMixin, models.Model),
        ),
        migrations.CreateModel(
            name="SubjectTime",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "day_of_week",
                    models.IntegerField(
                        choices=[
                            (1, "Monday"),
                            (2, "Tuesday"),
                            (3, "Wednesday"),
                            (4, "Thursday"),
                            (5, "Friday"),
                            (6, "Saturday"),
                            (7, "Sunday"),
                        ],
                        verbose_name="day of week",
                    ),
                ),
                ("start", models.TimeField(blank=True, null=True, verbose_name="start time")),
                ("end", models.TimeField(blank=True, null=True, verbose_name="end time")),
            ],
            options={
                "verbose_name": "time",
                "verbose_name_plural": "times",
                "ordering": ("day_of_week", "start"),
            },
            bases=(leprikon.models.startend.StartEndMixin, models.Model),
        ),
        migrations.AddField(
            model_name="subjecttime",
            name="subject",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="times",
                to="leprikon.Subject",
                verbose_name="course",
            ),
        ),
        migrations.AddField(
            model_name="journal",
            name="subject",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="journals",
                to="leprikon.Subject",
                verbose_name="subject",
            ),
        ),
        migrations.AddField(
            model_name="journalentry",
            name="journal",
            field=models.ForeignKey(
                default=None,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="journal_entries",
                to="leprikon.Journal",
                verbose_name="journal",
            ),
            preserve_default=False,
        ),
        migrations.RunSQL(
            sql="INSERT INTO leprikon_subjecttime SELECT * FROM leprikon_coursetime;",
            reverse_sql="INSERT INTO leprikon_coursetime SELECT * FROM leprikon_subjecttime;",
        ),
        migrations.RemoveField(
            model_name="coursetime",
            name="course",
        ),
        migrations.DeleteModel(
            name="CourseTime",
        ),
        migrations.RunPython(migrate_journals),
    ]



# Generated by Django 3.2.25 on 2025-04-10 11:37

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import leprikon.models.fields


def create_group_contacts(apps, schema_editor):
    RegistrationGroup = apps.get_model("leprikon", "RegistrationGroup")
    GroupContact = apps.get_model("leprikon", "GroupContact")

    group_contacts = {
        (g["user_id"], g["name"], g["first_name"], g["last_name"]): g
        for g in RegistrationGroup.objects.annotate(user_id=models.F("registration__user_id")).values(
            "user_id",
            "target_group_id",
            "name",
            "first_name",
            "last_name",
            "street",
            "city",
            "postal_code",
            "phone",
            "email",
            "school_id",
            "school_other",
            "school_class",
            "answers",
        )
    }

    GroupContact.objects.bulk_create(GroupContact(**group_contact) for group_contact in group_contacts.values())


def set_participants_count(apps, schema_editor):
    Registration = apps.get_model("leprikon", "Registration")

    for reg in Registration.objects.filter(activity__registration_type="P"):
        reg.participants_count = reg.participants.count()
        reg.save()

    for reg in Registration.objects.filter(activity__registration_type="G"):
        reg.participants_count = reg.group_members.count()
        reg.save()


def migrate_days_of_week(apps, schema_editor):
    for model_name in ("ActivityTime", "JournalTime"):
        Model = apps.get_model("leprikon", model_name)
        for day in leprikon.models.fields.DayOfWeek:
            Model.objects.filter(day_of_week=day.isoweekday()).update(
                days_of_week=leprikon.models.fields.DaysOfWeek(day)
            )


def migrate_orderable_registrations(apps, schema_editor):
    OrderableRegistration = apps.get_model("leprikon", "OrderableRegistration")
    CalendarEvent = apps.get_model("leprikon", "CalendarEvent")

    for registration in OrderableRegistration.objects.annotate(
        duration=django.db.models.F("activity__orderable__duration")
    ).iterator():
        start_date: datetime.date = registration.start_date
        start_time: datetime.time = registration.start_time or datetime.time(0, 0)
        duration: datetime.timedelta = registration.duration

        if start_time is None:
            end_date = start_date + duration
            end_time = None
        else:
            end = datetime.datetime.combine(start_date, start_time) + duration
            end_date = end.date()
            end_time = end.time()

        registration.calendar_event = CalendarEvent.objects.create(
            name=registration.activity.name,
            activity=registration.activity,
            start_date=start_date,
            start_time=start_time,
            end_date=end_date,
            end_time=end_time,
            is_canceled=registration.canceled is not None,
        )
        registration.save()


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("leprikon", "0088_activities"),
    ]

    operations = (
        [
            # Add GroupContact model
            migrations.CreateModel(
                name="GroupContact",
                fields=[
                    ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                    ("name", models.CharField(blank=True, default="", max_length=150, verbose_name="group name")),
                    ("first_name", models.CharField(max_length=30, verbose_name="first name")),
                    ("last_name", models.CharField(max_length=30, verbose_name="last name")),
                    ("street", models.CharField(max_length=150, verbose_name="street")),
                    ("city", models.CharField(max_length=150, verbose_name="city")),
                    ("postal_code", leprikon.models.fields.PostalCodeField(verbose_name="postal code")),
                    ("phone", models.CharField(max_length=30, verbose_name="phone")),
                    ("email", leprikon.models.fields.EmailField(max_length=254, verbose_name="email address")),
                    (
                        "school_other",
                        models.CharField(blank=True, default="", max_length=150, verbose_name="other school"),
                    ),
                    ("school_class", models.CharField(blank=True, default="", max_length=30, verbose_name="class")),
                    (
                        "answers",
                        models.TextField(blank=True, default="{}", editable=False, verbose_name="additional answers"),
                    ),
                    (
                        "school",
                        models.ForeignKey(
                            blank=True,
                            null=True,
                            on_delete=django.db.models.deletion.PROTECT,
                            related_name="+",
                            to="leprikon.school",
                            verbose_name="school",
                        ),
                    ),
                    (
                        "target_group",
                        models.ForeignKey(
                            on_delete=django.db.models.deletion.PROTECT,
                            related_name="+",
                            to="leprikon.targetgroup",
                            verbose_name="target group",
                        ),
                    ),
                    (
                        "user",
                        models.ForeignKey(
                            on_delete=django.db.models.deletion.CASCADE,
                            related_name="leprikon_group_contacts",
                            to=settings.AUTH_USER_MODEL,
                            verbose_name="user",
                        ),
                    ),
                ],
                options={
                    "verbose_name": "Group contact",
                    "verbose_name_plural": "Group contacts",
                },
            ),
            migrations.RunPython(create_group_contacts, reverse_code=migrations.RunPython.noop),
            # Add participants_count to Registration
            migrations.AddField(
                model_name="registration",
                name="participants_count",
                field=models.PositiveIntegerField(default=0, verbose_name="participants count"),
                preserve_default=False,
            ),
            migrations.RunPython(set_participants_count, reverse_code=migrations.RunPython.noop),
            # Migrate JournalTime and ActivityTime models to use DaysOfWeekField
        ]
        + [
            migrations.RenameField(
                model_name=model_name,
                old_name=field,
                new_name=f"{field}_time",
            )
            for model_name in ("journaltime", "activitytime")
            for field in ("start", "end")
        ]
        + [
            migrations.AddField(
                model_name="journaltime",
                name="days_of_week",
                field=leprikon.models.fields.DaysOfWeekField(default=0, verbose_name="days of week"),
                preserve_default=False,
            ),
            migrations.AddField(
                model_name="journaltime",
                name="end_date",
                field=models.DateField(
                    blank=True,
                    help_text="If not set, the time is valid for all dates.",
                    null=True,
                    verbose_name="end date",
                ),
            ),
            migrations.AddField(
                model_name="journaltime",
                name="start_date",
                field=models.DateField(
                    blank=True,
                    help_text="If not set, the time is valid for all dates.",
                    null=True,
                    verbose_name="start date",
                ),
            ),
            migrations.AddField(
                model_name="activitytime",
                name="days_of_week",
                field=leprikon.models.fields.DaysOfWeekField(default=0, verbose_name="days of week"),
                preserve_default=False,
            ),
            migrations.AddField(
                model_name="activitytime",
                name="end_date",
                field=models.DateField(
                    blank=True,
                    help_text="If not set, the time is valid for all dates.",
                    null=True,
                    verbose_name="end date",
                ),
            ),
            migrations.AddField(
                model_name="activitytime",
                name="start_date",
                field=models.DateField(
                    blank=True,
                    help_text="If not set, the time is valid for all dates.",
                    null=True,
                    verbose_name="start date",
                ),
            ),
            migrations.AlterField(
                model_name="journaltime",
                name="end_time",
                field=models.TimeField(
                    blank=True,
                    help_text="If end time is 0:00:00, it means the end of the day (effectively 24:00:00).",
                    null=True,
                    verbose_name="end time",
                ),
            ),
            migrations.AlterField(
                model_name="activitytime",
                name="end_time",
                field=models.TimeField(
                    blank=True,
                    help_text="If end time is 0:00:00, it means the end of the day (effectively 24:00:00).",
                    null=True,
                    verbose_name="end time",
                ),
            ),
            migrations.AlterModelOptions(
                name="journaltime",
                options={
                    "ordering": ("start_date", "days_of_week", "start_time"),
                    "verbose_name": "time",
                    "verbose_name_plural": "times",
                },
            ),
            migrations.AlterModelOptions(
                name="activitytime",
                options={
                    "ordering": ("start_date", "days_of_week", "start_time"),
                    "verbose_name": "time",
                    "verbose_name_plural": "times",
                },
            ),
            migrations.RunPython(migrate_days_of_week),
            # Add Resources
            migrations.CreateModel(
                name="Resource",
                fields=[
                    ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                    ("name", models.CharField(max_length=255, verbose_name="name")),
                    (
                        "leader",
                        models.OneToOneField(
                            blank=True,
                            null=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            related_name="resource",
                            to="leprikon.leader",
                            verbose_name="leader",
                        ),
                    ),
                ],
                options={
                    "verbose_name": "resource",
                    "verbose_name_plural": "resources",
                },
            ),
            migrations.CreateModel(
                name="ResourceAvailability",
                fields=[
                    ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                    ("start_date", models.DateField(blank=True, null=True, verbose_name="start date")),
                    ("end_date", models.DateField(blank=True, null=True, verbose_name="end date")),
                    ("days_of_week", leprikon.models.fields.DaysOfWeekField(verbose_name="days of week")),
                    ("start_time", models.TimeField(default=datetime.time(8, 0), verbose_name="start time")),
                    ("end_time", models.TimeField(default=datetime.time(16, 0), verbose_name="end time")),
                    (
                        "resource",
                        models.ForeignKey(
                            on_delete=django.db.models.deletion.CASCADE,
                            related_name="availabilities",
                            to="leprikon.resource",
                            verbose_name="resource",
                        ),
                    ),
                ],
                options={
                    "verbose_name": "availability",
                    "verbose_name_plural": "availabilities",
                    "ordering": ("start_time",),
                },
            ),
            migrations.CreateModel(
                name="ResourceGroup",
                fields=[
                    ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                    ("name", models.CharField(max_length=255, verbose_name="name")),
                    (
                        "resources",
                        models.ManyToManyField(related_name="groups", to="leprikon.Resource", verbose_name="resources"),
                    ),
                ],
                options={
                    "verbose_name": "resource group",
                    "verbose_name_plural": "resource groups",
                },
            ),
            # Migrate OrderableRegistration to use CalendarEvent
            migrations.CreateModel(
                name="CalendarEvent",
                fields=[
                    ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                    ("name", models.CharField(max_length=255, verbose_name="name")),
                    ("start_date", models.DateField(verbose_name="start date")),
                    ("start_time", models.TimeField(blank=True, null=True, verbose_name="start time")),
                    ("end_date", models.DateField(verbose_name="end date")),
                    ("end_time", models.TimeField(blank=True, null=True, verbose_name="end time")),
                    (
                        "preparation_time",
                        models.DurationField(
                            default=datetime.timedelta(0),
                            help_text="Time to prepare before the event. (HH:MM:SS)",
                            verbose_name="preparation time",
                        ),
                    ),
                    (
                        "recovery_time",
                        models.DurationField(
                            default=datetime.timedelta(0),
                            help_text="Time to recover after the event. (HH:MM:SS)",
                            verbose_name="recovery time",
                        ),
                    ),
                    ("is_canceled", models.BooleanField(default=False, verbose_name="is canceled")),
                    (
                        "resources",
                        models.ManyToManyField(
                            blank=True, related_name="calendar_events", to="leprikon.Resource", verbose_name="resources"
                        ),
                    ),
                    (
                        "resource_groups",
                        models.ManyToManyField(
                            blank=True,
                            related_name="calendar_events",
                            to="leprikon.ResourceGroup",
                            verbose_name="resource groups",
                        ),
                    ),
                    (
                        "activity",
                        models.ForeignKey(
                            blank=True,
                            null=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            related_name="calendar_events",
                            to="leprikon.activity",
                            verbose_name="activity",
                        ),
                    ),
                ],
                options={
                    "verbose_name": "calendar event",
                    "verbose_name_plural": "calendar events",
                    "ordering": ("start_date", "start_time"),
                },
            ),
            migrations.AddField(
                model_name="registration",
                name="calendar_event",
                field=models.OneToOneField(
                    null=True,
                    on_delete=django.db.models.deletion.PROTECT,
                    related_name="registration",
                    to="leprikon.calendarevent",
                    verbose_name="calendar event",
                ),
            ),
            migrations.AddField(
                model_name="activityvariant",
                name="require_group_members_list",
                field=models.BooleanField(
                    default=True,  # defaults to True for existing activity variants only
                    help_text="Require group members list to be filled in registration form. If not set, only the number of group members is required.",
                    verbose_name="require group members list",
                ),
            ),
            migrations.AlterField(
                model_name="activityvariant",
                name="require_group_members_list",
                field=models.BooleanField(
                    default=False,
                    help_text="Require group members list to be filled in registration form. If not set, only the number of group members is required.",
                    verbose_name="require group members list",
                ),
            ),
            migrations.AddField(
                model_name="activityvariant",
                name="required_resources",
                field=models.ManyToManyField(
                    blank=True,
                    related_name="activity_variants",
                    to="leprikon.Resource",
                    verbose_name="required resources",
                ),
            ),
            migrations.AddField(
                model_name="activityvariant",
                name="required_resource_groups",
                field=models.ManyToManyField(
                    blank=True,
                    related_name="activity_variants",
                    to="leprikon.ResourceGroup",
                    verbose_name="required resource groups",
                ),
            ),
            migrations.RunPython(migrate_orderable_registrations),
            migrations.AddField(
                model_name="orderable",
                name="preparation_time",
                field=models.DurationField(
                    default=datetime.timedelta(0),
                    help_text="Time to prepare before the event. (HH:MM:SS)",
                    verbose_name="preparation time",
                ),
            ),
            migrations.AddField(
                model_name="orderable",
                name="recovery_time",
                field=models.DurationField(
                    default=datetime.timedelta(0),
                    help_text="Time to recover after the event. (HH:MM:SS)",
                    verbose_name="recovery time",
                ),
            ),
            migrations.RemoveField(model_name="journaltime", name="day_of_week"),
            migrations.RemoveField(model_name="orderableregistration", name="start_date"),
            migrations.RemoveField(model_name="orderableregistration", name="start_time"),
            migrations.RemoveField(model_name="activitytime", name="day_of_week"),
        ]
    )



from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from .activities import ActivityType, ActivityVariant
from .leprikonsite import LeprikonSite
from .schoolyear import SchoolYear


class RegistrationLink(models.Model):
    school_year = models.ForeignKey(
        SchoolYear,
        editable=False,
        on_delete=models.CASCADE,
        related_name="registration_links",
        verbose_name=_("school year"),
    )
    activity_type = models.ForeignKey(
        ActivityType, on_delete=models.CASCADE, related_name="registration_links", verbose_name=_("activity type")
    )
    slug = models.SlugField(editable=False, max_length=64)
    name = models.CharField(_("name"), max_length=50)
    reg_from = models.DateTimeField(_("registration active from"))
    reg_to = models.DateTimeField(_("registration active to"))
    activity_variants = models.ManyToManyField(
        ActivityVariant, blank=False, related_name="registration_links", verbose_name=_("activities")
    )

    class Meta:
        app_label = "leprikon"
        ordering = ("reg_from", "reg_to")
        verbose_name = _("registration link")
        verbose_name_plural = _("registration links")

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse("leprikon:registration_link", args=(self.slug,))

    @cached_property
    def link(self):
        return LeprikonSite.objects.get_current().url + self.get_absolute_url()

    @property
    def registration_allowed(self):
        now = timezone.now()
        return (self.reg_from <= now) and (self.reg_to > now)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-03-21 12:57
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0048_birth_date"),
    ]

    operations = [
        migrations.AddField(
            model_name="courselistplugin",
            name="target_groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by target groups.",
                related_name="+",
                to="leprikon.TargetGroup",
                verbose_name="target groups",
            ),
        ),
        migrations.AddField(
            model_name="eventlistplugin",
            name="target_groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by target groups.",
                related_name="+",
                to="leprikon.TargetGroup",
                verbose_name="target groups",
            ),
        ),
    ]
